#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
"""
Here is a sample code for a system test.
The purpose of the test is to test the tag's transmission characteristics during different scenarios.
The script can be run from the command line or from a GUI (by running it on any Python platform).

The test parameters are stored in the 'system_test_menu_defaults_file.json' file.
 to run the script using command line:
        go to the script location and type: system_test_example.py -f <file_name>
 <file_name> is the test parameters file, e.g.: system_test_menu_defaults_file.json

Outputs from tests are stored under the outputs folder.
Each test is saved under a different folder with an increasing number.

Change the calibration parameters under get_setup_calibration_data according to your setup.

Review all '# TODO' in the script and change/edit/add according to instructions
"""
import json
import serial.tools.list_ports
import argparse
import tkinter as tk
import tkinter.font as font
import tkinter.filedialog
import matplotlib.pyplot as plt
import numpy as np
import csv
from wiliot_core import *

# Set Global Parameters:
plt.switch_backend('agg')  # clean and init plot
verbose_flag = False
menu_defaults_global = {'verbose_flag': 1, 'tag_distance': 2.5, 'positioner_enable': 0, 'tag_angle': 0,
                        'gw_att_type': 'None', 'user_gw_dca_raw': 0, 'test_time_sec': 20.0, 'pause_time_sec': 180.0,
                        'number_of_test_iterations': 3, 'GW_Baud': 921600,
                        'energizing_pattern_list': '20', 'gw_timing_list': '(5/15)', 'N_LTBP': 10,
                        'pos_type_var': 'None', 'gw_port': 'Other', 'gw_att_port_var': 'Other',
                        'gw_port_other': ' ', 'gw_att_port_entry': ' ', 'num_rx_packets_to_stop_test': 'NA'}


def save_to_csv(output_csv_file_name, output_csv_file_fields, out_rows):
    """
    save data to the specified csv file

    :type output_csv_file_name: str
    :param output_csv_file_name: the file name where the data is going to be stored
    :type output_csv_file_fields: list
    :param output_csv_file_fields: the headers of the data to be stored
    :type out_rows: np.array
    :param out_rows: the data to be stored

    :return
    """
    with open(output_csv_file_name, 'w', newline='') as csv_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerow(output_csv_file_fields)
        csv_writer.writerows(out_rows)


def extract_port_selection(port_var, port_entry):
    """
    extract serial port for connection

    :type port_var: str
    :param port_var: the port variable according to the drop-down menu (all available ports + 'Other' option)
    :type port_entry: str
    :param port_entry: the manually added port. relevant only when port_var is 'Other'

    :return the relevant port
    """
    if 'Other' in port_var:
        port = port_entry
        print('Manual serial port selection: "{}"'.format(port))
    else:
        port = port_var
    
    return port


def extract_value_from_user_entry(user_entry, is_range=True):
    """
    extract the test parameters from a string when a string can contain a range, e.g. 0:15:90
    or a list of option, e.g. 20,17,52

    :type user_entry: str
    :param user_entry: the user entry string
    :type is_range: bool
    :param is_range: True if user input is a range of number, False if it is a list of options

    :return a list of the extracted value
    """
    if ':' in user_entry:
        sep_str = ':'
    elif ',' in user_entry:
        sep_str = ','
    elif '/' in user_entry:
        sep_str = '/'
    else:
        try:
            if '.' in user_entry:
                value_out = [float(user_entry)]
            else:
                value_out = [int(user_entry)]
            return value_out
        except Exception as e:
            print('please enter range in the following format 0:1:10 or 0,1,10 or '
                  'list in the following format 20,17,52 or 5/15')
            raise SystemExit
    
    try:
        value_int = [int(n) for n in user_entry.split(sep_str)]
    except ValueError:
        try:
            value_int = [float(n) for n in user_entry.split(sep_str)]
        except Exception as e:
            print('!' * 10 + ' Error: Could Not Extract Range from ' + user_entry + '!' * 10)
            print(e)
            raise SystemExit
    if is_range:
        if len(value_int) != 3:
            print('please enter range in the following format 0:1:10 or 0,1,10 or '
                  'list in the following format 20,17,52 or 5/15')
            raise SystemExit
        value_out = np.arange(value_int[0], value_int[2] + value_int[1], value_int[1])
    else:
        value_out = value_int
    return value_out


def extract_dca_att_range(user_gw_dca_raw):
    """
    extract the DCA Attenuator parameters from a string when a string can contain a range, e.g. 0:0.25:2
    or a number, e.g. 0

    :type user_gw_dca_raw: str
    :param user_gw_dca_raw: the user entry attenuation string

    :return a list of the extracted attenuation
    """
    dca_att_range_raw = extract_value_from_user_entry(user_gw_dca_raw, is_range=True)
    dca_att_range = []
    for dca_val in dca_att_range_raw:
        # TODO add here validation test for your DCA Attenuator:
        fixed_dca = dca_val
        if system_test_hardware.dca_valid_range is not None:
            fixed_dca = max(min(dca_val, system_test_hardware.dca_valid_range[1]),
                            system_test_hardware.dca_valid_range[0])  # make sure the user value is in range
        if dca_val != fixed_dca:
            print('?' * 10 + ' User-Defined DCA Value (' + str(dca_val) + 'dB) changed to: ' + str(
                fixed_dca) + 'dB')
        
        dca_att_range.append(fixed_dca)
    
    dca_att_range = np.unique(dca_att_range)
    return dca_att_range


def extract_positioner_angle_range(tag_angle, positioner_enable):
    """
    extract the positioner parameters from a string when a string can contain the angle range (deg), e.g. 0:15:90
    or a number (deg), e.g. 0

    :type positioner_enable: bool
    :param positioner_enable: the user entry position (angles) string
    :type tag_angle: str
    :param tag_angle: the user entry attenuation string

    :return a list of the extracted angles
    """
    if positioner_enable == 0:
        positioner_angle_range = [0]
    else:
        positioner_angle_range = extract_value_from_user_entry(tag_angle, is_range=True)
    return positioner_angle_range


def extract_gw_timing_list(gw_timing_str):
    """
    extract the time profiles parameters from a string when a string can contain a list of profiles, e.g. (5/15),(0/8)
    or a single profile, e.g. (5/15)

    :type gw_timing_str: str
    :param gw_timing_str: the user entry time profile string

    :return a list of the extracted time profiles
    """
    try:
        start_i = [i for i, c in enumerate(gw_timing_str) if c == '(']
        stop_i = [i for i, c in enumerate(gw_timing_str) if c == ')']
        time_profile_str = [gw_timing_str[i + 1:j] for i, j in zip(start_i, stop_i)]
        time_profile_list = [extract_value_from_user_entry(tp, is_range=False) for tp in time_profile_str]
        
        return time_profile_list
    except Exception as e:
        print('Error: Could not Extract GW Timing List!!! please write in the following format:(x/x),(x/x)')
        raise SystemExit


def generate_high_level_box_plot(results_data_mat, meas_type, y_label, data_col, number_of_test_iterations):
    """
    plot a boxplot of a specific measurement (meas_type, data_col) over all iteration per each test parameter.
    Each scenario (a set of test parameters) is plotted in a different subplot and each tag has a different box.
    The statistics is more significant as the number of iteration increases.
    The graph is saved as a pdf file and stored under the test folder

    :type results_data_mat: np.matrix
    :param results_data_mat: the
    :type meas_type: str
    :param meas_type: the specific measurement name to display
    :type y_label: str
    :param y_label: the y label string
    :type data_col: int
    :param data_col: the column of the relevant data (the specific measurements) to display
    :type number_of_test_iterations: int
    :param number_of_test_iterations: the number of iterations for each test

    :return a list of the extracted time profiles
    """
    tag_id_vec = results_data_mat[:, 9]
    angles_vec = results_data_mat[:, 6].astype(float)
    atts_vec = results_data_mat[:, 16].astype(float)
    energizing_pattern_vec = results_data_mat[:, 11].astype(int)
    time_profile_vec = results_data_mat[:, 12].astype(str)
    
    # per each parameter (attenuation, angles, timing profiles and energizing patterns):
    energizing_pattern_uni = np.unique(np.array(energizing_pattern_vec).flatten())
    time_profile_uni = np.unique(np.array(time_profile_vec).flatten())
    tag_id_uni = np.unique(np.array(tag_id_vec).flatten())
    angles_uni = np.unique(np.array(angles_vec).flatten())
    atts_uni = np.unique(np.array(atts_vec).flatten())
    
    n_rows = int(np.ceil(np.sqrt(angles_uni.shape[0])))
    n_cols = int(np.ceil(np.sqrt(angles_uni.shape[0])))
    
    output_graph_file_name = ''
    
    if angles_uni.shape[0] <= 9:
        font_size = 14
    else:
        font_size = 10
    
    for att in atts_uni:
        try:
            att = int(att)
        except Exception as e:
            pass
        for energizing_pattern in energizing_pattern_uni:
            for time_profile in time_profile_uni:
                if np.argwhere(np.logical_and(energizing_pattern_vec == energizing_pattern,
                                              np.logical_and(time_profile_vec == time_profile,
                                                             atts_vec == att)))[:, 0].shape[0] == 0:
                    continue
                
                plt.figure(figsize=(36, 28), dpi=35)
                for ang_index, ang in enumerate(angles_uni):
                    try:
                        ang = int(ang)
                    except Exception as e:
                        pass
                    output_sub_case_folder, _ = \
                        system_test_log.set_output_folder_and_files(ang, att,
                                                                    extract_gw_timing_list('(' + time_profile.
                                                                                           replace('ms', '') + ')')[0],
                                                                    energizing_pattern)
                    output_graph_file_name = os.path.join(output_sub_case_folder,
                                                          '{} - all iterations.pdf'.format(meas_type))
                    values = []
                    tag_enumeration = ''
                    x_labels = []
                    for tag_index, tag_id in enumerate(tag_id_uni):
                        tag_enumeration = tag_enumeration + ' ' + str(tag_index + 1) + '=' + tag_id + '\n'
                        value_indices = np.argwhere(np.logical_and(
                            np.logical_and(energizing_pattern_vec == energizing_pattern,
                                           time_profile_vec == time_profile),
                            np.logical_and(angles_vec == ang, tag_id_vec == tag_id)))[:, 0]
                        responses_vec = results_data_mat[value_indices, 14].astype(int)
                        values.append(
                            np.array(results_data_mat[value_indices, data_col]).astype(float).flatten())
                        x_labels.append(str(tag_index + 1) + '\n' + str(np.sum(responses_vec)) + '/' + str(
                            number_of_test_iterations))
                    
                    plt.subplot(n_rows, n_cols, ang_index + 1)
                    
                    plt.boxplot(values, vert=True, patch_artist=True, labels=x_labels)
                    plt.rcParams.update({'font.size': font_size - 2})
                    plt.title('ATT=' + str(att) + 'dB, Angle=' + str(int(ang)) + '$^o$', fontsize=font_size)
                    plt.xlabel('Tag Index\\n/N', fontsize=font_size)
                    plt.ylabel(y_label, fontsize=font_size)
                    plt.grid()
                    
                    if ang_index + 1 == angles_uni.shape[0]:
                        bottom_x, top_x = plt.xlim()
                        bottom_y, top_y = plt.ylim()
                        plt.text(top_x, bottom_y, tag_enumeration, fontsize=font_size - 2)
                
                plt.suptitle('Test #' + str(system_test_log.max_folder_number) + ': Energizing Pattern=' + str(
                    energizing_pattern) + ', Timing Profile=' + time_profile + ', ' + str(
                    tag_id_uni.shape[0]) + ' Tags', fontsize=24)
                
                plt.savefig(output_graph_file_name)
                plt.close('all')
                if verbose_flag:
                    print(output_graph_file_name)
        
        plt.close('all')


def init_test_parameters():
    """
    initialize the test parameters by constant values
    :return a dictionary of the test parameters
    """
    test_parameters = {}
    system_test_log.set_main_test_folder()
    test_parameters['max_folder_number'] = system_test_log.max_folder_number
    test_parameters['main_test_folder'] = system_test_log.main_test_folder
    # ignoring the long time of the first packet transmission:
    test_parameters['num_packets_for_max_time_between_packets'] = 1
    test_parameters['output_csv_file_fields'] = ['Time to First Packet [sec]', 'Number of Received Packets (GW)',
                                                 'RX Rate (GW) [Hz]', 'Normalized RX Rate (GW) [Hz]',
                                                 'Longest Time between Packets (GW) [sec] ',
                                                 'Longest Time between Packets after ' +
                                                 str(test_parameters['num_packets_for_max_time_between_packets']) +
                                                 ' Packets (GW) [sec]',
                                                 'Tag Angle [Deg]', 'Tag Distance [m]', 'Test Iteration', 'Tag ID',
                                                 'GW BLE Chip SW Version', 'Energizing Pattern', 'Timing Profile',
                                                 'Test Time [sec]', 'Response', 'ATT Type', 'ATT Setting [dB]',
                                                 'RSSI',
                                                 'Energizing Emulated Distance [m]',
                                                 'Beacons Emulated Distance [m]']
    test_parameters['test_time_sec'] = 0.0
    test_parameters['pause_time_sec'] = 0.0
    return test_parameters


def cli_config_to_test_param(config_data):
    """
    according to the configuration file content and to constant values,
    the function assigns the relevant values to the test parameters and return them

    :type config_data: dict
    :param config_data: the configuration data

    :return a dictionary of the test parameters
    """
    test_parameters = init_test_parameters()
    global verbose_flag
    verbose_flag = config_data['verbose_flag']
    for key, value in config_data.items():
        test_parameters[key] = value
    # arrange unique parameters:
    test_parameters['energizing_pattern_list'] = extract_value_from_user_entry(config_data['energizing_pattern_list'],
                                                                               is_range=False)
    test_parameters['gw_timing_list'] = extract_gw_timing_list(str(config_data['gw_timing_list']))
    test_parameters['test_time_sec'] = float(config_data['test_time_sec'])
    test_parameters['pause_time_sec'] = float(config_data['pause_time_sec'])
    num_rx_packets_to_stop_test = config_data['num_rx_packets_to_stop_test']
    try:
        test_parameters['num_rx_packets_to_stop_test'] = int(num_rx_packets_to_stop_test)
    except Exception as e:
        test_parameters['num_rx_packets_to_stop_test'] = 0
    
    return test_parameters


class Log(object):
    """
    This class is responsible for all the files and folders name
    """
    
    def __init__(self):
        
        self.outputs_folder = os.path.join(os.path.dirname(__file__), 'outputs')  # outputs folder name
        self.max_folder_number = 0
        self.main_test_folder = ''
        self.config_file_name = ''
    
    def set_main_test_folder(self):
        """
        set the main test folder name.
        For each test a new folder is generated anf its name is an increases number
        """
        if not os.path.exists(self.outputs_folder):
            os.makedirs(self.outputs_folder)
        else:
            last_test_run_folders = os.listdir(self.outputs_folder)
            last_test_run_num = []
            for folder_name in last_test_run_folders:
                try:
                    last_test_run_num.append(int(folder_name))
                except Exception as e:
                    pass
            if last_test_run_num:
                self.max_folder_number = max(last_test_run_num)
        
        self.max_folder_number = self.max_folder_number + 1
        
        self.main_test_folder = os.path.join(self.outputs_folder, str(self.max_folder_number))
        if not os.path.exists(self.main_test_folder):
            os.makedirs(self.main_test_folder)
            if verbose_flag:
                print('Created Main Output Folder: {}'.format(self.main_test_folder))
        else:
            print('!' * 10 + ' ' + self.main_test_folder + ' already exist' + '!' * 10)
            raise SystemExit
        
        return
    
    def set_output_folder_and_files(self, positioner_angle, actual_gw_dca_att, time_profile, energizing_pattern):
        """
        set the output folders for each scenario, i.e.e set of test parameters according to the function inputs
        :type positioner_angle: int or str
        :param positioner_angle: the current test positioner angle in deg
        :type actual_gw_dca_att: int or str
        :param actual_gw_dca_att: the current test attenuation in dB
        :type time_profile: list
        :param time_profile: the current test GW time profile [on-value, period] in ms
        :type energizing_pattern: int or str
        :param energizing_pattern: the current test GW energizing pattern
        """
        output_sub_case_folder = os.path.join(self.main_test_folder, 'Ang=' + str(positioner_angle) + ' Deg',
                                              'ATT=' + str(actual_gw_dca_att) + 'dB',
                                              'Timing=(' + str(time_profile[0]) + ',' + str(time_profile[1]) + ')',
                                              'Pattern=' + str(energizing_pattern))
        
        if not os.path.exists(output_sub_case_folder):
            os.makedirs(output_sub_case_folder)
            if verbose_flag:
                print('Created Output Folder: ' + output_sub_case_folder)
        
        results_output_csv_file_name = os.path.join(output_sub_case_folder,
                                                    str(self.max_folder_number) + '. MTRST - Results.csv')
        return output_sub_case_folder, results_output_csv_file_name
    
    def set_config_file_name(self, new_file_name):
        """
        set new name to the configuration file

        :type new_file_name str
        :param new_file_name the new config file name
        """
        self.config_file_name = new_file_name


class Config(object):
    """
    This class is responsible for all the configuration functions
    """
    
    def __init__(self):
        self.calibration_param = []
    
    def get_setup_calibration_data(self):
        """
        GW_Pout - GW power out (in dBm)
        Setup_Conducted_Insertion_Line - the signal loss due to cable and connectors
        Horn_Gain - the signal/loss gain due to the antenna used in the test
        f_transmit - the GW freq in GHz
        Ref_GW_EIRP - the reference GW Effective Radiated Power (the estimated GW power out with its antenna
        """
        # TODO Add here your setup calibration parameters
        # CAL_Params=[GW_Pout, Setup_Conducted_IL, Horn_Gain, f_transmit, Ref_GW_EIRP]
        self.calibration_param = [21.7, 3.4, 11.4, 2.48, 23]
        return self.calibration_param
    
    def calc_emulated_distance(self, gw_att, tag_distance):
        """
        This function calculates the estimated emulated distance
        or the actual distance between the GW and the tag in the open space

        According to Friis equation [in dB]:
        transmitter = GW
        receiver = tag

        according to the test setup: Pr = Pt + Gt + Gr + 20*log(lambda/(4*pi*d))
        Pt = gw_power_out - setup_conducted_insertion_line + horn_gain [dB]
        d = tag_distance
        Gt = 0 [dB], Gr = 0 [dB]

        estimating the real-life-application distance between gw and tag in open space:
        d_e = (10^-((Pr_e - Pt_e -Gt_e - Gr_e)/20)) * (lambda/(4*pi))
        Pr_e = Pr [dB]
        Pt_e = reference_gw_EIRP [dB]
        Gt_e, Gr_e = 0 [dB]

        lambda - wavelength = v/f, when v in free space 3*10^8 and f = f_transmit*10^9
        lambda = (3*10^8)/(f_transmit*10^9) = 0.3/f_transmit
        """
        gw_pout, setup_conducted_il, horn_gain, f_transmit, ref_gw_eirp = self.calibration_param
        dis_tag = float(tag_distance)
        beacons_backoff = 0  # gw config parameter for reducing the beacons power. default is 0
        
        estimated_distance = []
        calc_types = ['energizing', 'beacons']
        for dis_type in calc_types:
            lambda_wave = 0.3 / f_transmit  # wave length according to the gw frequency in GHz
            p_t_setup = 0  # init
            if dis_type == 'energizing':
                p_t_setup = gw_pout - setup_conducted_il + horn_gain - gw_att  # the setup gain and loss components
            elif dis_type == 'beacons':
                p_t_setup = gw_pout - setup_conducted_il + horn_gain - gw_att - beacons_backoff
            
            p_r = p_t_setup + 20 * np.log10(lambda_wave / (4 * np.pi * dis_tag))  # the power available at the receiver
            # calculate the estimated distance between the gw's antenna and tag's antenna without the setup's components
            d_e = round((10 ** (-((p_r - ref_gw_eirp) / 20))) * (lambda_wave / (4 * np.pi)), 2)
            estimated_distance.append(d_e)
        
        return estimated_distance[0], estimated_distance[1]


class GUI(object):
    """
    This class is responsible for all the GUI functions
    """
    
    def __init__(self):
        """
        initialize all parameters and GUI screen
        """
        # init default values:
        self.stop_all = False
        self.system_test_menu_defaults_file_name = os.path.join(os.path.dirname(__file__),
                                                                'system_test_menu_defaults_file.json')
        if not os.path.isfile(self.system_test_menu_defaults_file_name) or \
                os.stat(self.system_test_menu_defaults_file_name).st_size == 0:
            # save the default values to json
            with open(self.system_test_menu_defaults_file_name, "w") as f:
                json.dump(menu_defaults_global, f)
        
        # start the GUI application
        master = tk.Tk()
        master.geometry("600x800+0+0")
        master.title("Multi-Tag System Test")
        font.nametofont('TkDefaultFont').configure(size=12)
        
        self.pad_x_0 = 10
        self.pad_x_1 = 350
        
        self.verbose_enable_checkbutton_status = tk.IntVar()
        tk.Checkbutton(master, text="Verbose Mode", variable=self.verbose_enable_checkbutton_status). \
            grid(row=0, padx=self.pad_x_0, sticky=tk.W)
        
        tk.Label(master, text="Distance [m]", fg='black').grid(row=1, column=0, sticky=tk.W, padx=self.pad_x_1 * .73)
        
        positioner_enable_checkbutton_status = tk.IntVar()
        tk.Checkbutton(master, text="Enable Positioner", variable=positioner_enable_checkbutton_status, fg='red'). \
            grid(row=2, padx=self.pad_x_0, sticky=tk.W)
        tk.Label(master, text="Positioner Type", fg='red').grid(row=2, column=0, sticky=tk.W, padx=self.pad_x_1 * .65)
        
        pos_type_list = ['Positioner1', 'Positioner2', 'Positioner3']
        pos_type_var = tk.StringVar()
        tk.OptionMenu(master, pos_type_var, *pos_type_list).grid(row=2, padx=self.pad_x_1, sticky=tk.W)
        
        tk.Label(master, text="Tag Angle [Degrees]", fg='red').grid(row=3, column=0, sticky=tk.W, padx=self.pad_x_0)
        tk.Label(master, text="GW ATT Type", fg='magenta').grid(row=4, padx=self.pad_x_0, sticky=tk.W)
        
        dca_type_list = ['DCA1', 'DCA2', 'None']
        self.gw_dca_type_var = tk.StringVar()
        tk.OptionMenu(master, self.gw_dca_type_var, *dca_type_list).grid(row=4, padx=self.pad_x_1, sticky=tk.W)
        
        tk.Label(master, text="GW ATT Value [dB]", fg='magenta').grid(row=5, padx=self.pad_x_0, sticky=tk.W)
        tk.Label(master, text="Test Time [sec]", fg='red').grid(row=9, column=0, sticky=tk.W, padx=self.pad_x_0)
        tk.Label(master, text="Pause Time [sec]", fg='red').grid(row=10, column=0, sticky=tk.W, padx=self.pad_x_0)
        tk.Label(master, text="Number of Iterations", fg='red').grid(row=11, column=0, sticky=tk.W, padx=self.pad_x_0)
        tk.Label(master, text="GW Energizing Pattern", fg='magenta').grid(row=14, column=0, sticky=tk.W,
                                                                          padx=self.pad_x_0)
        tk.Label(master, text="GW Timing Profile (ON-Time/Period) [ms]", fg='magenta').grid(row=15, column=0,
                                                                                            sticky=tk.W,
                                                                                            padx=self.pad_x_0)
        tk.Label(master, text="Required Number of RX Packets to Stop Test", fg='black').grid(row=19, column=0,
                                                                                             sticky=tk.W,
                                                                                             padx=self.pad_x_0)
        
        distance_entry = tk.Entry(master, fg='black', font="Calibri 12")
        tag_angle_entry = tk.Entry(master, fg='red', font="Calibri 12")
        gw_att_entry = tk.Entry(master, fg='magenta', font="Calibri 12")
        test_time_entry = tk.Entry(master, fg='red', font="Calibri 12")
        pause_time_entry = tk.Entry(master, fg='red', font="Calibri 12")
        num_iterations_entry = tk.Entry(master, fg='red', font="Calibri 12")
        energizing_pattern_entry = tk.Entry(master, fg='magenta', font="Calibri 12")
        timing_profile_entry = tk.Entry(master, fg='magenta', font="Calibri 12")
        num_packet_to_stop_single_tag_test_entry = tk.Entry(master, fg='black', font="Calibri 12")
        
        distance_entry.grid(row=1, column=0, sticky=tk.W, padx=self.pad_x_1)
        tag_angle_entry.grid(row=3, column=0, sticky=tk.W, padx=self.pad_x_1)
        gw_att_entry.grid(row=5, padx=self.pad_x_1, sticky=tk.W)
        test_time_entry.grid(row=9, column=0, sticky=tk.W, padx=self.pad_x_1)
        pause_time_entry.grid(row=10, column=0, sticky=tk.W, padx=self.pad_x_1)
        num_iterations_entry.grid(row=11, column=0, sticky=tk.W, padx=self.pad_x_1)
        energizing_pattern_entry.grid(row=14, column=0, sticky=tk.W, padx=self.pad_x_1)
        timing_profile_entry.grid(row=15, column=0, sticky=tk.W, padx=self.pad_x_1)
        num_packet_to_stop_single_tag_test_entry.grid(row=19, column=0, sticky=tk.W, padx=self.pad_x_1)
        
        tk.Button(master, text='Set Defaults', command=self.set_defaults, bg='blue', fg='yellow'). \
            grid(row=20, column=0, sticky=tk.W, padx=self.pad_x_1, pady=5)
        
        self.runtime_label = tk.Label(master, text='', fg='black')
        self.runtime_label.grid(row=21, column=0, sticky=tk.W, padx=180, pady=5)
        tk.Button(master, text='Calculate Runtime', command=self.calculate_runtime, bg='white', fg='green'). \
            grid(row=21, column=0, sticky=tk.W, padx=10, pady=5)
        
        self.config_file_name_label = tk.Label(master, text='', fg='black')
        self.config_file_name_label.grid(row=23, column=0, sticky=tk.W, padx=210)
        tk.Button(master, text='Set Config File Name', command=self.gui_set_config_file_name, bg='blue', fg='white'). \
            grid(row=23, column=0, sticky=tk.W, padx=10)
        tk.Button(master, text='Generate Config File', command=self.gui_generate_config_file,
                  bg='yellow', fg='black').grid(row=24, column=0, sticky=tk.W, pady=5, padx=10)
        
        available_ports = [s.device for s in serial.tools.list_ports.comports()]
        if len(available_ports) == 0:
            available_ports = [s.name for s in serial.tools.list_ports.comports()]
        available_ports.append('Other')
        # assign to gui variables
        si_labs_ports = available_ports
        non_si_labs_port = available_ports
        
        gw_port_var = tk.StringVar()
        tk.Label(master, text="GW Port", fg='black').grid(row=25, padx=self.pad_x_0, sticky=tk.W)
        tk.OptionMenu(master, gw_port_var, *si_labs_ports).grid(row=25, padx=self.pad_x_0 + 150, sticky=tk.W)
        
        gw_port_entry = tk.Entry(master, fg='black', font="Calibri 12")
        gw_port_entry.grid(row=25, column=0, sticky=tk.W, padx=self.pad_x_0 + 250)
        
        tk.Label(master, text="GW ATT Port", fg='black').grid(row=27, padx=self.pad_x_0, sticky=tk.W)
        gw_att_port_var = tk.StringVar()
        tk.OptionMenu(master, gw_att_port_var, *non_si_labs_port).grid(row=27, padx=self.pad_x_0 + 150, sticky=tk.W)
        
        gw_att_port_entry = tk.Entry(master, fg='black', font="Calibri 12")
        gw_att_port_entry.grid(row=27, column=0, sticky=tk.W, padx=self.pad_x_0 + 250)
        
        tk.Button(master, text='Run Test', command=lambda: run_main(gui_obj=self), bg='green', fg='white', width=20,
                  height=1,
                  font="Calibri 16").grid(row=29, column=0, sticky=tk.W, padx=10, pady=5)
        tk.Button(master, text='Quit', command=self.quit_gui, bg='red', fg='white', width=20, height=1,
                  font="Calibri 16").grid(row=29, column=0, sticky=tk.W, padx=250, pady=5)
        
        master.protocol("WM_DELETE_WINDOW", self.quit_gui)
        # saved variables:
        self.master = master
        self.gw_port_var = gw_port_var
        self.gw_port_entry = gw_port_entry
        self.gw_att_port_var = gw_att_port_var
        self.gw_att_port_entry = gw_att_port_entry
        self.pos_type_var = pos_type_var
        self.gw_att_entry = gw_att_entry
        self.tag_angle_entry = tag_angle_entry
        self.positioner_enable_checkbutton_status = positioner_enable_checkbutton_status
        self.distance_entry = distance_entry
        self.test_time_entry = test_time_entry
        self.pause_time_entry = pause_time_entry
        self.num_iterations_entry = num_iterations_entry
        self.energizing_pattern_entry = energizing_pattern_entry
        self.timing_profile_entry = timing_profile_entry
        self.num_packet_to_stop_single_tag_test_entry = num_packet_to_stop_single_tag_test_entry
        # set gui fields:
        try:
            self.revert_selections()
        except Exception as e:
            self.set_defaults()
        # start
        self.master.mainloop()
    
    def gui_extract_test_parameters(self, file_name_to_save=None):
        """
        according to the GUI user entry content and to constant values,
        the function assigns the relevant values to the test parameters and return them

        :type file_name_to_save: str
        :param file_name_to_save: the file name which the test parameters will be stored.
                                  if None the test parameters are stored at the default file (see Log() class)

        :return a dictionary of the test parameters
        """
        test_parameters = init_test_parameters()
        global verbose_flag
        verbose_flag = self.verbose_enable_checkbutton_status.get()
        test_parameters['verbose_flag'] = verbose_flag
        test_parameters['gw_att_type'] = self.gw_dca_type_var.get()
        test_parameters['gw_att_port_var'] = self.gw_att_port_var.get()
        test_parameters['gw_att_port_entry'] = self.gw_att_port_entry.get()
        test_parameters['user_gw_dca_raw'] = str(self.gw_att_entry.get())
        test_parameters['tag_angle'] = str(self.tag_angle_entry.get())
        test_parameters['positioner_enable'] = self.positioner_enable_checkbutton_status.get()
        test_parameters['pos_type_var'] = self.pos_type_var.get()
        test_parameters['tag_distance'] = str(self.distance_entry.get())
        test_parameters['test_time_sec'] = float(self.test_time_entry.get())
        test_parameters['pause_time_sec'] = float(self.pause_time_entry.get())
        test_parameters['number_of_test_iterations'] = int(self.num_iterations_entry.get())
        test_parameters['energizing_pattern_list'] = \
            extract_value_from_user_entry(self.energizing_pattern_entry.get(), is_range=False)
        test_parameters['gw_timing_list'] = extract_gw_timing_list(str(self.timing_profile_entry.get()))
        
        num_rx_packets_to_stop_test = self.num_packet_to_stop_single_tag_test_entry.get()
        if num_rx_packets_to_stop_test.isnumeric():
            test_parameters['num_rx_packets_to_stop_test'] = int(num_rx_packets_to_stop_test)
        else:
            test_parameters['num_rx_packets_to_stop_test'] = 0
        
        test_parameters['gw_port'] = extract_port_selection(self.gw_port_var.get(), self.gw_port_entry.get())
        
        menu_defaults = {'verbose_flag': self.verbose_enable_checkbutton_status.get(),
                         'tag_distance': self.distance_entry.get(),
                         'positioner_enable': self.positioner_enable_checkbutton_status.get(),
                         'tag_angle': self.tag_angle_entry.get(),
                         'gw_att_type': self.gw_dca_type_var.get(),
                         'user_gw_dca_raw': self.gw_att_entry.get(),
                         'test_time_sec': self.test_time_entry.get(),
                         'pause_time_sec': test_parameters['pause_time_sec'],
                         'number_of_test_iterations': test_parameters['number_of_test_iterations'],
                         'energizing_pattern_list': self.energizing_pattern_entry.get(),
                         'gw_timing_list': self.timing_profile_entry.get(),
                         'pos_type_var': self.pos_type_var.get(),
                         'gw_port': self.gw_port_var.get(),
                         'gw_att_port_var': self.gw_att_port_var.get(),
                         'gw_port_other': self.gw_port_entry.get(),
                         'gw_att_port_entry': self.gw_att_port_entry.get(),
                         'num_rx_packets_to_stop_test': test_parameters['num_rx_packets_to_stop_test']}
        if file_name_to_save is None:
            file_name_to_save = self.system_test_menu_defaults_file_name
        json.dump(menu_defaults, open(file_name_to_save, "w"))
        
        return test_parameters
    
    def gui_set_config_file_name(self):
        """
        starts to run if 'set config file' button is pressed.
        If a name was not selected after he dialog pop-up a default name to the config file is generated
        """
        # open a dialog
        user_selection = tk.filedialog.asksaveasfile(initialdir=os.path.dirname(__file__),
                                                     filetypes=[('json files', '*.json')],
                                                     defaultextension='.json')
        
        if user_selection is not None:
            system_test_log.set_config_file_name(user_selection.name)
        else:  # if the user did not select a name, a default name is created
            now = datetime.datetime.now()
            timestamp = datetime.datetime.timestamp(now)
            timestamp_str = str(timestamp).replace('.', '_')
            system_test_log.set_config_file_name(os.path.dirname(__file__) + '/config_' + timestamp_str + '.json')
        
        self.config_file_name_label.destroy()
        self.config_file_name_label = tk.Label(self.master, text=system_test_log.config_file_name,
                                               fg='black', wraplength=350, justify='left', font="Calibri 8")
        self.config_file_name_label.grid(row=23, column=0, sticky=tk.W, padx=210)
    
    def gui_generate_config_file(self):
        """
        starts to run if 'generate config file' button is pressed.
        this function saves all test parameters to the configuration file
        """
        if is_gui:  # this function runs only when gui is enabled
            now = datetime.datetime.now()
            timestamp = datetime.datetime.timestamp(now)
            timestamp_str = str(timestamp).replace('.', '_')
            if not system_test_log.config_file_name:
                system_test_log.set_config_file_name(
                    os.path.join(os.path.dirname(__file__), 'config_{}.json'.format(timestamp_str)))
            try:
                self.gui_extract_test_parameters(file_name_to_save=system_test_log.config_file_name)
                print('Configuration saved to file: ' + system_test_log.config_file_name)
            except Exception as e:
                print('an error occurred during saving new configuration:\n{}'.format(e))
        else:
            print("gui_generate_config_file is running although Command Line run method was applied")
            return
        
        tk.Label(self.master, text=system_test_log.config_file_name + ' saved' + ' ' * 1000, fg='black', wraplength=350,
                 justify='left', font="Calibri 8").grid(row=24, column=0, sticky=tk.W, padx=200)
    
    def calculate_runtime(self):
        """
        starts to run if 'calculate run time' button is pressed.
        this function calculate the whole test run time according to all user entry parameters including
        parameters range and pauses between tests iterations/cycles
        """
        tag_angle = str(self.tag_angle_entry.get())
        positioner_enable = self.positioner_enable_checkbutton_status.get()
        positioner_angle_range = extract_positioner_angle_range(tag_angle, positioner_enable)
        gw_att_type = self.gw_dca_type_var.get()
        user_gw_dca_raw = str(self.gw_att_entry.get())
        if gw_att_type == 'None':
            print('DCA Attenuator type was not define, hence no attenuation range is set')
            num_atts = 1
        else:
            dca_att_range = extract_dca_att_range(user_gw_dca_raw)
            num_atts = len(dca_att_range)
        
        if positioner_enable == 0:
            num_angles = 1
        else:
            num_angles = len(positioner_angle_range)
        
        test_time_sec = float(self.test_time_entry.get())
        pause_time_sec = float(self.pause_time_entry.get())
        number_of_test_iterations = int(self.num_iterations_entry.get())
        energizing_pattern_list = list(int(x) for x in self.energizing_pattern_entry.get().split(','))
        gw_timing_str = str(self.timing_profile_entry.get())
        gw_timing_list = extract_gw_timing_list(gw_timing_str)
        
        length_gw_timing_list = np.shape(gw_timing_list)[0]
        # calculate the number of iterations
        overall_number_of_iterations = num_atts * num_angles * number_of_test_iterations * len(
            energizing_pattern_list) * length_gw_timing_list
        # calculate the test tine in seconds:
        estimated_runtime_sec = overall_number_of_iterations * (test_time_sec + pause_time_sec + 5)
        
        if estimated_runtime_sec > 60:
            estimated_runtime_string = str(round(estimated_runtime_sec / 60, 2)) + ' Minutes'
        else:
            estimated_runtime_string = str(round(estimated_runtime_sec, 2)) + ' Seconds'
        if estimated_runtime_sec > 3600:
            estimated_runtime_string = str(round(estimated_runtime_sec / 3600, 2)) + ' Hours'
        if estimated_runtime_sec > 86400:
            estimated_runtime_string = str(round(estimated_runtime_sec / 86400, 2)) + ' Days'
        
        self.runtime_label.destroy()
        self.runtime_label = tk.Label(self.master, text=estimated_runtime_string, fg='black')
        self.runtime_label.grid(row=21, column=0, sticky=tk.W, padx=180, pady=5)
    
    def quit_gui(self):
        """
        starts to run if 'quit' button is pressed.
        This function can be used to stop all processes in a safe way
        """
        print('Operation terminated by user.')
        self.stop_all = True  # a flag that terminates a test even if not all iterations are completed
        
        raise SystemExit
    
    def set_gui_fields(self, menu_defaults):
        """
        This function sets all GUI fields and assign a default values

        :type menu_defaults: dict
        :param menu_defaults: the GUI fields default values
        """
        self.distance_entry.delete(0, 'end')
        self.tag_angle_entry.delete(0, 'end')
        self.test_time_entry.delete(0, 'end')
        self.pause_time_entry.delete(0, 'end')
        self.num_iterations_entry.delete(0, 'end')
        self.energizing_pattern_entry.delete(0, 'end')
        self.timing_profile_entry.delete(0, 'end')
        self.num_packet_to_stop_single_tag_test_entry.delete(0, 'end')
        self.gw_att_entry.delete(0, 'end')
        self.gw_port_entry.delete(0, 'end')
        self.gw_att_port_entry.delete(0, 'end')
        self.distance_entry.insert(10, menu_defaults['tag_distance'])
        self.tag_angle_entry.insert(10, menu_defaults['tag_angle'])
        self.test_time_entry.insert(10, menu_defaults['test_time_sec'])
        self.pause_time_entry.insert(10, menu_defaults['pause_time_sec'])
        self.num_iterations_entry.insert(10, menu_defaults['number_of_test_iterations'])
        self.energizing_pattern_entry.insert(10, menu_defaults['energizing_pattern_list'])
        self.timing_profile_entry.insert(10, menu_defaults['gw_timing_list'])
        self.num_packet_to_stop_single_tag_test_entry.insert(10, menu_defaults['num_rx_packets_to_stop_test'])
        self.gw_att_entry.insert(10, menu_defaults['user_gw_dca_raw'])
        
        self.positioner_enable_checkbutton_status.set(int(menu_defaults['positioner_enable']))
        self.pos_type_var.set(str(menu_defaults['pos_type_var']))
        self.verbose_enable_checkbutton_status.set(int(menu_defaults['verbose_flag']))
        self.gw_dca_type_var.set(str(menu_defaults['gw_att_type']))
        self.gw_port_var.set(menu_defaults['gw_port'])
        self.gw_att_port_var.set(menu_defaults['gw_att_port_var'])
        self.gw_att_port_entry.insert(10, menu_defaults['gw_att_port_entry'])
        
        self.calculate_runtime()
    
    def set_defaults(self):
        """
        Sets the default values for all GUI fields
        """
        menu_defaults = menu_defaults_global
        self.set_gui_fields(menu_defaults)
    
    def revert_selections(self):
        """
        starts to run if 'reset to default' button is pressed.
        This function overwrite users changes in the GUI and set the default values instead
        """
        try:
            menu_defaults = json.load(open(self.system_test_menu_defaults_file_name, "rb"))
        except Exception as e:
            menu_defaults = menu_defaults_global
            print('?' * 10 + ' Incompatible History File - Using Defaults Instead.')
        
        self.set_gui_fields(menu_defaults)


class TestEquipment(object):
    """
    This class is responsible for all scripts related to the test equipments including external position
    and DCA attenuator, that can be implemented during the system test.
    This class is only an example since for each setup a different functions and interfaces should be implemented
    """
    
    def __init__(self):
        """
        initialize equipment
        """
        # TODO add here specific range or conditions for each DCA type, e.g. valid range:
        self.dca_valid_range = [0, 20]
        # TODO add here specific range or conditions for the positioner, e.g. valid range:
    
    def test_equipment_connect_to_positioner(self, positioner_enable, pos_type_var, tag_angle):
        """
        This function connects to the positioner and return the positioner class and connections status

        :type positioner_enable bool
        :param positioner_enable: specified if the positioner is enable in the test setup
        :type pos_type_var str
        :param pos_type_var: the positioner name/type
        :type tag_angle str
        :param tag_angle: the desired angles for the positioner during the test

        :return the connection status (bool), the positioner object, the angles range (list of angles in deg)
        """
        if positioner_enable == 1:
            print("Connect to positioner: {}".format(pos_type_var))
            # TODO add here a connection to your positioner
            connect_pos_flag = True  # if connection was establish, connect_pos_flag is True
            my_pos = []  # your positioner class for further actions
            pos_range = extract_positioner_angle_range(tag_angle,
                                                       positioner_enable)  # the test range for the positioner
        else:
            connect_pos_flag = False
            my_pos = None
            pos_range = None
        return connect_pos_flag, my_pos, pos_range
    
    def test_equipment_connect_to_dca(self, att_type, port_var, port_entry, user_dca_raw):
        """
        This function connects to the DCA attenuator and return the attenuator class and connections status

        :type att_type str
        :param att_type: the attenuator name/type
        :type port_var str
        :param port_var: the attenuator serial port
        :type port_entry str
        :param port_entry: the attenuator manually serial port (relevant only if port_var is 'Other'
        :type user_dca_raw str
        :param user_dca_raw: the desired attenuation range for the positioner during the test

        :return the connection status (bool), the positioner object, the angles range (list of angles in deg)
        """
        dca_port = extract_port_selection(port_var, port_entry)
        print("Connect to DCA: type {}, port:".format(att_type, dca_port))
        # TODO add here a connection to your DCA logic
        connect_dca_flag = True  # if connection was establish, connect_dac_flag is True
        my_att = []  # your attenuator class for further actions
        dca_att_range = extract_dca_att_range(user_dca_raw)  # the test range for the attenuator
        
        return connect_dca_flag, my_att, dca_att_range
    
    def exit(self):
        """
        This function exit and close all processes related to the test equipment
        """
        # TODO close all connections and exit
        pass


# ======================Main Code Starts===============================================


def run_test(gui_obj, output_sub_case_folder, gw_obj, test_time_sec, test_iteration, energizing_pattern,
             num_packets_for_max_time_between_packets,
             tag_angle_str, tag_distance, time_profile, gw_att_type, actual_gw_dca_att,
             gw_ver, num_rx_packets_to_stop_test, t_start):
    """
    This function runs the core test
    First parameters and GW are initiated
    Then a while loop runs until timeout or after N received packets if num_rx_packets_to_stop_test is specified
    During the while loop the tags' packets are stored and checked for validity and for new tags
    At the end of the loop the GW is configured to stop transmitting energy to the tags.
    After each iteration and csv with all packets data is saved together with two graphs (as pdf files) containing the
    boxplot (distribution) of the time between packets and a temporal graph of the number of packets per tag vs. time

    :type gui_obj: GUI class
    :param gui_obj: the object of the GUI class
    :type output_sub_case_folder: str
    :param output_sub_case_folder: the folder path of the current test results
    :type gw_obj: WiliotGateway
    :param gw_obj: the object of the WiliotGateway class
    :type test_time_sec: int
    :param test_time_sec: the test timeout in seconds
    :type test_iteration: int
    :param test_iteration: the current test iteration
    :type energizing_pattern: int
    :param energizing_pattern: the current test energizing pattern for the GW
    :type num_packets_for_max_time_between_packets: int
    :param num_packets_for_max_time_between_packets: the number of packet from which the longest time after N packets
                                                    is measured
    :type tag_angle_str: str
    :param tag_angle_str: the tag angle in deg
    :type tag_distance: int or float
    :param tag_distance: the tag distance in the test setup
    :type time_profile: list
    :param time_profile: the current test time profile for the GW
    :type gw_att_type: str
    :param gw_att_type: the attenuator type for logging
    :type actual_gw_dca_att: int or float
    :param actual_gw_dca_att: the current test attentuation
    :type gw_ver: str
    :param gw_ver: the GW version for logging
    :type num_rx_packets_to_stop_test: int
    :param num_rx_packets_to_stop_test: if specified (larger than 0) the current test continued until
                                        the received packets have reached this number or if timeout occurred
    :type t_start: int
    :param t_start: the start time of the test in seconds

    :return the data output (2D list), the test completion time in sec
    """
    dis_energizing, dis_beacons = system_test_config.calc_emulated_distance(actual_gw_dca_att, tag_distance)
    test_time_sec_local = test_time_sec
    
    packet_data = []
    t_vec = []
    tag_id_vec = []
    rssi_vec = []
    tag_id_list = []
    
    time_profile_str = str(time_profile[0]) + 'ms/' + str(time_profile[1]) + 'ms'
    # GW:
    if gw_obj is None:
        gw_obj = WiliotGateway(baud=921600, auto_connect=True, verbose=False)
    # config
    gw_obj.config_gw(energy_pattern_val=energizing_pattern, time_profile_val=time_profile)
    gw_obj.check_current_config()
    gw_obj.reset_buffer()
    
    t0 = time.time()
    timeout = t0 + test_time_sec_local
    
    print('Starting Multi-Tag System Test (Duration=' + str(test_time_sec_local) + ' Seconds)...')
    time_remaining_prev = test_time_sec_local
    gw_obj.start_continuous_listener()
    
    while time.time() < timeout:
        time.sleep(0.0)  # help the cpu to recover between loops
        if gui_obj is not None:
            gui_obj.master.update()
            if gui_obj.stop_all:
                break
        if not gw_obj.is_data_available():
            continue
        # data is available - let's get data:
        data_in = gw_obj.get_packets(action_type=ActionType.FIRST_SAMPLES, num_of_packets=1,
                                     data_type=DataType.PROCESSED)
        if not data_in:
            continue
        data_in = data_in[0]
        # new packet has arrived:
        packet_data.append(data_in['packet'])
        t_vec.append(data_in['time_from_start'])
        # check the advertising address:
        tag_id_vec.append(data_in['adv_address'])
        rssi_vec.append(data_in['rssi'])
        if not data_in['adv_address'] in tag_id_list:
            print('********** New Tag Detected: ' + data_in['adv_address'] + ' **********')
            tag_id_list.append(data_in['adv_address'])
        
        # check if need to print:
        time_remaining = timeout - time.time()
        if time_remaining_prev - time_remaining >= 1:
            print(tag_angle_str + u'\N{DEGREE SIGN}, GW ATT [dB] ' + str(actual_gw_dca_att) + ' (' + str(
                dis_energizing) + ' m), ' + time_profile_str +
                  ', Pattern ' + str(energizing_pattern) + ', Iteration ' + str(test_iteration) +
                  ', Time Remaining [sec]: ' + str(round(time_remaining)) + ', Tag Count: ' + str(len(tag_id_list)) +
                  ', Overall Packet Count: ' + str(len(packet_data)))
            
            time_remaining_prev = time_remaining
        
        # check if we received enough packets:
        if num_rx_packets_to_stop_test > 0:
            if len(packet_data) >= num_rx_packets_to_stop_test:
                print('*' * 10 + ' ' + str(
                    num_rx_packets_to_stop_test) + ' Packets Received - Stopping Iteration. ' + '*' * 10)
                test_time_sec_local = t_vec[-1]
                break
    
    gw_obj.stop_continuous_listener()
    gw_obj.config_gw(time_profile_val=[0, 15])  # stop transmitting
    gw_obj.check_current_config()
    
    # check if there are more packets in the GW API buffer that have received before the end of the test:
    while True:
        data_in = gw_obj.get_packets(action_type=ActionType.FIRST_SAMPLES, num_of_packets=1,
                                     data_type=DataType.PROCESSED)
        if not data_in or data_in[0]['time_from_start'] > test_time_sec_local:
            break
        else:
            data_in = data_in[0]
            packet_data.append(data_in['packet'])
            t_vec.append(data_in['time_from_start'])
            tag_id_vec.append(data_in['adv_address'])
            if not (data_in['adv_address'] in tag_id_list):
                print('********** New Tag Detected: ' + data_in['adv_address'] + ' **********')
                tag_id_list.append(data_in['adv_address'])
    
    # write data to csv:
    if len(packet_data) > 0:
        packets_output_csv_file_name = output_sub_case_folder + '/' + str(test_iteration) + '. MTRST - Packets Data.csv'
        output_csv_file_fields = ['Time [sec]', 'Tag ID', 'Packets Data']
        save_to_csv(packets_output_csv_file_name, output_csv_file_fields,
                    np.transpose([t_vec, tag_id_vec, packet_data]))
    
    # check duplication
    packet_data_wo_rssi = []
    for full_packet in packet_data:
        packet_data_wo_rssi.append(full_packet[:74])
    
    _, unique_packet_data_indices = np.unique(packet_data_wo_rssi,
                                              return_index=True)
    unique_packet_data_indices = np.sort(unique_packet_data_indices)
    unique_packet_data = np.array(packet_data)[unique_packet_data_indices.astype(int)]
    t_vec_without_duplicated_packets = np.array(t_vec)[unique_packet_data_indices.astype(int)]
    tag_id_vec_without_duplicated_packets = np.array(tag_id_vec)[unique_packet_data_indices.astype(int)]
    tag_id_list = np.unique(tag_id_vec_without_duplicated_packets)
    
    # arrange data to save csv and graphs (as pdf)
    output_wo_dup = []
    
    for packet_index in range(len(unique_packet_data)):
        pixie_analyzer_out_row = [t_start + t_vec[packet_index], tag_angle_str, tag_distance, test_iteration, gw_ver,
                                  energizing_pattern, time_profile_str, test_time_sec_local, gw_att_type,
                                  actual_gw_dca_att, dis_energizing, dis_beacons]
        
        output_wo_dup.append(pixie_analyzer_out_row)
    
    plt.figure(figsize=(16, 12), dpi=100)
    temporal_graph_file_name = output_sub_case_folder + '/' + str(
        test_iteration) + '. MTRST - Temporal Graph.pdf'
    box_plot_tbp_file_name = output_sub_case_folder + '/' + str(test_iteration) + '. MTRST - TBP BP.pdf'
    
    out_rows = []
    rates_rx_tx = []
    ttfp_vec = []
    tbp = []
    
    for tag_id, rssi in zip(tag_id_list, rssi_vec):
        # find the indices of each tag (w/o duplications)
        indices_wo_dup = np.argwhere(np.array(tag_id_vec_without_duplicated_packets).flatten() == tag_id)
        t_wo_dup = np.array(t_vec_without_duplicated_packets[indices_wo_dup]).flatten()
        # calculate the time between packets
        tbp.append(list(np.diff(t_wo_dup)))
        # find the indices of each tag (including duplications)
        indices_with_dup = np.argwhere(np.array(tag_id_vec).flatten() == tag_id)
        t_with_dup = np.array(np.array(t_vec)[indices_with_dup]).flatten()
        # find the time to first packet of each tag
        ttfp = t_wo_dup[0]
        ttfp_vec.append(ttfp)
        num_rx_packets = len(indices_wo_dup)  # number of packets per tag
        # calculate the receiving rate: #of packets divided by th time
        rx_rate = num_rx_packets / test_time_sec_local
        # calculate the longest time between packets
        if num_rx_packets == 1:
            ltbp = test_time_sec_local - t_wo_dup[-1]
            norm_rx_rate = 1 / test_time_sec_local
            ltbp_after_n_packets = 0
        else:
            ltbp = np.max([np.max(np.diff(t_wo_dup)), test_time_sec_local - t_wo_dup[
                -1]])
            if test_time_sec_local - ttfp == 0:  # not supposed to happen
                norm_rx_rate = num_rx_packets / test_time_sec_local
            else:
                norm_rx_rate = num_rx_packets / (test_time_sec_local - ttfp)
            if num_rx_packets > num_packets_for_max_time_between_packets + 1:
                # longest time between packets after n Packets
                ltbp_after_n_packets = np.max(
                    [np.max(np.diff(t_wo_dup[num_packets_for_max_time_between_packets - 1:-1])),
                     test_time_sec_local - t_wo_dup[-1]])
            else:
                ltbp_after_n_packets = 0
        
        # plotting the received packets vs. time
        p = plt.step(t_with_dup, range(1, len(t_with_dup) + 1), '--', where='pre', label=None)
        plt.step(t_wo_dup, range(1, len(t_wo_dup) + 1), ':', color=p[-1].get_color(), label=tag_id + ', RX')
        
        out_row = [np.transpose(np.array(
            [ttfp, num_rx_packets, rx_rate, norm_rx_rate, ltbp, ltbp_after_n_packets, tag_angle_str, tag_distance,
             test_iteration, tag_id, gw_ver, energizing_pattern, time_profile_str, test_time_sec_local, '1',
             gw_att_type, actual_gw_dca_att, rssi, dis_energizing, dis_beacons]))]
        # concatenate data:
        if len(out_rows) == 0:
            out_rows = out_row
        else:
            if len(out_row) > 0:
                out_rows = np.vstack([out_rows, out_row])
        
        rates_rx_tx.append(norm_rx_rate)
    
    if len(tag_id_list) > 0:
        plt.legend()
        plt.title('Tag Count: ' + str(len(tag_id_list)))
        plt.xlabel('Time [sec]')
        plt.ylabel('Packet Count')
        plt.grid(axis='both')
        plt.savefig(temporal_graph_file_name)
        plt.close('all')
        
        plt.figure(figsize=(16, 12), dpi=100)
        plt.boxplot(tbp, vert=True, patch_artist=True, labels=tag_id_list)
        plt.title('Tag Count: ' + str(len(tag_id_list)))
        plt.ylabel('Time between Packets [sec]')
        plt.grid()
        plt.savefig(box_plot_tbp_file_name)
        plt.close('all')
    
    plt.close('all')
    
    return out_rows, test_time_sec_local


def run_main(with_gui=True, gui_obj=None):
    """
    This function run the whole test with all its iterations
    At first, the function initializes all parameters, establishes connection with the GW and all other test equipment,
    and sets all test folder names.
    Then it runs over several loops, a loop per each user-specified parameter.
    for each scenario, i.e. a different set of test parameter, the function calls the run_test to preform the test.
    At the end of each iteration a combined file containing all iteration results is saved and stored under the
    test folder. At the end of all iterations the function exits from all functions, GW and test equipment processes
    """
    system_test_config.get_setup_calibration_data()
    generate_high_level_box_plot_flag = False
    if with_gui:
        test_parameters = gui_obj.gui_extract_test_parameters()  # Not CLI
    else:
        if not cli_configuration:
            print("please add test parameters to '{}'".format(system_test_log.config_file_name))
            raise SystemExit
        else:
            test_parameters = cli_config_to_test_param(cli_configuration)
    
    number_of_test_iterations = test_parameters['number_of_test_iterations']
    
    # Gateway:
    gw_obj = WiliotGateway(baud=921600, port=test_parameters['gw_port'], verbose=False)
    gw_obj.reset_gw()
    # config to main param without transmitting (time profile - ON value is 0)
    gw_obj.config_gw(pacer_val=False, time_profile_val=[0, 15], beacons_backoff_val=0,
                     received_channel=37, modulation_val=True)
    gw_sw_ver, gw_hw_ver = gw_obj.get_gw_version()
    gw_ver = gw_hw_ver + '=' + gw_sw_ver
    # DCA Attenuator:
    connect_gw_dca_flag, my_gw_att, gw_dca_att_range = \
        system_test_hardware.test_equipment_connect_to_dca(test_parameters['gw_att_type'],
                                                           test_parameters['gw_att_port_var'],
                                                           test_parameters['gw_att_port_entry'],
                                                           test_parameters['user_gw_dca_raw'])
    # Positioner
    connect_pos_flag, my_pos, positioner_angle_range = \
        system_test_hardware.test_equipment_connect_to_positioner(bool(test_parameters['positioner_enable']),
                                                                  test_parameters['pos_type_var'],
                                                                  test_parameters['tag_angle'])
    
    output_sub_case_folder_list = []
    out_rows_all_test_cases = []
    
    t_start = 0
    
    for positioner_angle in positioner_angle_range:  # positioner loop
        if test_parameters['positioner_enable'] == 1:
            print('Moving Positioner by ' + str(positioner_angle) + ' Degrees.')
            # TODO implement positioner movement:
            # my_pos.move_to_pos_deg(positioner_angle)
            tag_angle_str = str(positioner_angle)
        else:
            tag_angle_str = test_parameters['tag_angle']
        
        for gw_dca_att in gw_dca_att_range:  # attenuation loop
            if connect_gw_dca_flag:
                print('DCA set to ' + str(gw_dca_att) + 'dB')
                # TODO implement attenuation setup for your DCA:
                # my_gw_att.set_attn(gw_dca_att)
                actual_gw_dca_att = gw_dca_att
            else:
                actual_gw_dca_att = 0
            
            for time_profile in test_parameters['gw_timing_list']:  # time profile loop
                for energizing_pattern in test_parameters['energizing_pattern_list']:  # energizing pattern loop
                    test_iteration = 0
                    out_rows_all = []
                    
                    while test_iteration < number_of_test_iterations:  # iterations loop
                        test_iteration = test_iteration + 1
                        if gui_obj is not None:
                            if gui_obj.stop_all:
                                break
                        output_sub_case_folder, results_output_csv_file_name = \
                            system_test_log.set_output_folder_and_files(positioner_angle, actual_gw_dca_att,
                                                                        time_profile, energizing_pattern)
                        output_sub_case_folder_list.append(output_sub_case_folder)
                        print('Iteration Number ' + str(test_iteration) + ', Pausing for ' + str(
                            test_parameters['pause_time_sec']) + ' Seconds...')
                        time.sleep(test_parameters['pause_time_sec'])
                        
                        t_start = t_start + test_parameters['pause_time_sec']
                        
                        out_rows, test_time_sec_local = \
                            run_test(gui_obj, output_sub_case_folder, gw_obj,
                                     test_parameters['test_time_sec'], test_iteration,
                                     energizing_pattern, test_parameters['num_packets_for_max_time_between_packets'],
                                     tag_angle_str, test_parameters['tag_distance'],
                                     time_profile, test_parameters['gw_att_type'], actual_gw_dca_att,
                                     gw_ver, test_parameters['num_rx_packets_to_stop_test'], t_start)
                        
                        t_start = t_start + test_time_sec_local
                        # concatenate results:
                        if len(out_rows) > 0:
                            if len(out_rows_all) == 0:
                                out_rows_all = out_rows
                            else:
                                if len(out_rows) > 0:
                                    out_rows_all = np.vstack([out_rows_all, out_rows])
                            
                            if verbose_flag:
                                # print(out_rows_all)
                                print(results_output_csv_file_name)
                            
                            save_to_csv(results_output_csv_file_name, test_parameters['output_csv_file_fields'],
                                        out_rows_all)
                    
                    if len(out_rows_all_test_cases) == 0:
                        out_rows_all_test_cases = out_rows_all
                    else:
                        if len(out_rows_all) > 0:
                            out_rows_all_test_cases = np.vstack((out_rows_all_test_cases, out_rows_all))
                    
                    if len(out_rows_all_test_cases) > 0:
                        combined_output_csv_file_name = os.path.join(system_test_log.main_test_folder,
                                                                     '{}. MTRST - All Results.csv'.
                                                                     format(system_test_log.max_folder_number))
                        save_to_csv(combined_output_csv_file_name, test_parameters['output_csv_file_fields'],
                                    out_rows_all_test_cases)
                        
                        # TODO change this flag to true if you want to generate box plot including
                        #  all iterations of a specific test parameters
                        if generate_high_level_box_plot_flag:
                            generate_high_level_box_plot(np.matrix(out_rows_all_test_cases), meas_type='TTFP',
                                                         y_label='TTFP [sec]', data_col=0,
                                                         number_of_test_iterations=number_of_test_iterations)
                            generate_high_level_box_plot(np.matrix(out_rows_all_test_cases), meas_type='RX-Rate',
                                                         y_label='Normalized RX Rate w/o Dup. [Hz]', data_col=2,
                                                         number_of_test_iterations=number_of_test_iterations)
                            generate_high_level_box_plot(np.matrix(out_rows_all_test_cases), meas_type='LTBP',
                                                         y_label='LTBP [sec]', data_col=4,
                                                         number_of_test_iterations=number_of_test_iterations)
                    if gui_obj is not None:
                        if gui_obj.stop_all:
                            gw_obj.exit_gw_api()
                            system_test_hardware.exit()
                            print('Done.')
                            return
    
    # close all connection
    gw_obj.exit_gw_api()
    system_test_hardware.exit()
    print('Done.')
    
    raise SystemExit


# ======================Main Code Ends===============================================
if __name__ == '__main__':
    # init Log and Config
    is_gui = False
    system_test_log = Log()
    system_test_config = Config()
    system_test_hardware = TestEquipment()
    # read from command line if applicable:
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store', dest='File_Name',
                        help='File_Name - Configuration File Name (Json Format, *.json)')
    args = vars(parser.parse_args())
    system_test_log.config_file_name = args['File_Name']
    cli_configuration = {}
    
    if system_test_log.config_file_name is None:
        # run using GUI
        is_gui = True
        system_test_gui = GUI()
    
    else:
        # run using Command line
        is_gui = False
        cli_configuration = json.load(open(system_test_log.config_file_name, "rb"))
        verbose_flag = cli_configuration['verbose_flag']
        run_main(with_gui=False)



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import logging
import os.path
import datetime
import csv
from wiliot_api import TesterType
from wiliot_core import WiliotDir
from wiliot_tools.utils.wiliot_gui.wiliot_gui import WiliotGui, popup_message

from wiliot_testers.utils.upload_to_cloud_api import upload_to_cloud_api
from wiliot_testers.wiliot_tester_tag_result import FailureCodes

MAX_FILE_SIZE = 15 * 10 ** 6  # 15 Mb


def create_summary_message(is_upload=True, data_folder_name=''):
    """
    summary log
    :param is_upload:
    :type is_upload: bool or None
    :return:
    :rtype:
    @param data_folder_name:
    @type data_folder_name:  str
    """
    sum_msg_prefix = '\n\n\n***********************************************\n'
    sum_msg_suffix = '***********************************************\n'
    sum_msg_upload = f'********* UPLOAD: files were uploaded from {data_folder_name} *********\n'
    sum_msg = sum_msg_prefix
    if is_upload is not None:
        sum_msg += sum_msg_upload
        if not is_upload:
            sum_msg = sum_msg.replace('were', 'were NOT')
    sum_msg += sum_msg_suffix
    return sum_msg


class UploadTestersData(object):
    def __init__(self):
        """
        This function check the selected user files and upload them to the cloud.
        If the file exceed the size limit it split it
        """
        self.values = None
        self.logger = None

        self.get_files_path()
        if self.values is None:
            print('user exited the program')
            return

        self.set_logging()
        file_status = self.check_files()
        if not file_status:
            raise Exception(f'invalid file inputs: {self.values["run_data_file"]}, {self.values["packets_data_file"]}')

        # check file size:
        self.all_run_names = [self.values['run_data_file']]
        self.all_packets_names = [self.values['packets_data_file']]
        self.tester_type = self.values['tester_type']
        file_size = os.stat(self.values['packets_data_file']).st_size
        if file_size > MAX_FILE_SIZE and 'offline' in self.tester_type.lower():
            self.split_large_file(run_data=self.values['run_data_file'], packet_data=self.values['packets_data_file'],
                                  size=file_size)

    def get_files_path(self):
        """
        opens GUI for selecting a file and returns it
        """

        layout_dict = {
            'run_data_file': {'text': 'Choose run data file that you want to upload:', 'value': '',
                              'widget_type': 'file_input'},
            'packets_data_file': {'text': 'Choose packets data file that you want to upload:', 'value': '',
                                  'widget_type': 'file_input'},
            'env': {'text': 'Environment', 'value': "prod", 'widget_type': 'combobox',
                    'options': ('prod', 'test', 'dev')},
            'owner_id': {'text': 'Owner id:', 'value': '852213717688', 'widget_type': 'entry'},
            'tester_type': {'text': 'Tester type:', 'value': "OFFLINE_TEST", 'widget_type': 'combobox',
                            'options': tuple(TesterType.__members__.keys())}}

        gui = WiliotGui(params_dict=layout_dict)
        values_out = gui.run()

        if values_out['run_data_file'] != '' or values_out['packets_data_file'] != '':
            self.values = values_out
        else:
            self.values = None

    def set_logging(self):
        """
        Sets up logging for the object, creating a log file with the current date and time as part of the filename.
        Returns: None
        """
        #  get()dirname from run_data or packets_data
        files_dir = os.path.dirname(self.values.get('run_data_file', self.values.get('packets_data_file')))
        if not files_dir:
            files_dir = os.path.join(WiliotDir().get_tester_dir("upload_to_cloud"), "logs")
            os.makedirs(files_dir, exist_ok=True)
        log_path = os.path.join(files_dir,
                                datetime.datetime.now().strftime("%d-%m-%Y_%H-%M-%S") + 'upload_to_cloud.log')
        logger = logging.getLogger('Manual Upload')
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s', '%H:%M:%S')
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(formatter)
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        stream_handler.setLevel(logging.DEBUG)
        logger.addHandler(file_handler)
        logger.addHandler(stream_handler)
        self.logger = logger

    def check_files(self):
        """
        Checks if the run_data_file and packets_data_file are valid CSV files.
        Returns True if both files are valid, False otherwise.
        """
        if not os.path.isfile(self.values['run_data_file']) or not self.values['run_data_file'].endswith('.csv'):
            self.logger.warning('UPLOAD: run_data file format is not csv, please insert a csv file')
            self.logger.info(create_summary_message(False))
            return False
        if not os.path.isfile(self.values['packets_data_file']) or not self.values['packets_data_file'].endswith(
                '.csv'):
            self.logger.warning('UPLOAD: packets_data file format is not csv, please insert a csv file')
            self.logger.info(create_summary_message(False))
            return False
        return True

    def split_large_file(self, run_data, packet_data, size):

        """
        Splits a large packet_data CSV file into smaller parts that are each under the maximum allowed size.
        Duplicates the corresponding run_data CSV file for each part. Returns a list of file paths for each split part.
        """

        # Define a helper function to generate new file paths.
        def get_new_path_name(original_path, k, total, file_str):
            new_folder_name = os.path.dirname(original_path) + f'_{k}_out_of_{n_splits}'
            new_file_name = os.path.basename(original_path).replace(f'@{file_str}_data.csv',
                                                                    f'_{k}_out_of_{total}@{file_str}_data.csv')
            if not os.path.isdir(new_folder_name):
                os.makedirs(new_folder_name)
            return os.path.join(new_folder_name, new_file_name)

        # Split the file into smaller parts.
        n_splits = -(-size // MAX_FILE_SIZE)
        part = 1
        all_packets_path = []
        all_run_path = []
        # read
        f_read = open(packet_data, 'r')
        reader = csv.DictReader(f_read, delimiter=',')
        col_names = reader.fieldnames
        # write first file
        new_packet_path = get_new_path_name(packet_data, part, n_splits, 'packets')
        all_packets_path.append(new_packet_path)
        f_write = open(new_packet_path, 'w', newline='')
        writer = csv.DictWriter(f_write, fieldnames=col_names)
        writer.writeheader()
        last_loc = 0
        tested_list = []
        start_loc = -1
        missing_label_count = 0
        for i, row in enumerate(reader):
            row['common_run_name'] = os.path.basename(new_packet_path).split('@')[0]
            if os.stat(new_packet_path).st_size < MAX_FILE_SIZE or int(row['tag_run_location']) == last_loc:
                writer.writerow(row)
                last_loc = int(row['tag_run_location'])
                missing_label_count += FailureCodes.MISSING_LABEL.value == int(row['fail_bin'])
            else:  # we need to split the file
                tested_list.append(last_loc - start_loc - missing_label_count)
                f_write.close()
                part += 1
                new_packet_path = get_new_path_name(packet_data, part, n_splits, 'packets')
                all_packets_path.append(new_packet_path)
                f_write = open(new_packet_path, 'w', newline='')
                writer = csv.DictWriter(f_write, fieldnames=col_names)
                writer.writeheader()
                row['common_run_name'] = os.path.basename(new_packet_path).split('@')[0]
                writer.writerow(row)
                start_loc = last_loc
                last_loc = int(row['tag_run_location'])
                missing_label_count = 0
        tested_list.append(last_loc - start_loc)
        f_write.close()
        if part > n_splits:
            self.logger.info(
                f'Since the file cannot be split in the middle of location data, a mismatch happened and '
                f'the number of parts is bigger than the number of split (i.e you will have a file with '
                f'the suufix {part}_out_of{n_splits}')
        # Duplicate the run data for each part.
        for i in range(part):
            new_run_path = get_new_path_name(run_data, i + 1, n_splits, 'run')
            all_run_path.append(new_run_path)
            with open(run_data, 'r') as f_run_read:
                run_reader = csv.DictReader(f_run_read, delimiter=',')
                run_col_names = run_reader.fieldnames
                run_row = run_reader.__next__()
            run_row['common_run_name'] = os.path.basename(new_run_path).split('@')[0]
            run_row['total_run_tested'] = str(tested_list[i])
            with open(new_run_path, 'w', newline='') as f_run_write:
                run_writer = csv.DictWriter(f_run_write, fieldnames=run_col_names)
                run_writer.writeheader()
                run_writer.writerow(run_row)

        self.all_run_names = all_run_path
        self.all_packets_names = all_packets_path

    def upload_to_cloud(self):
        """
        Uploads files to a cloud API
        Returns True if all uploads were successful
        """
        all_status = []
        if self.tester_type.upper() not in TesterType.__members__.keys():
            raise Exception('unsupported tester type, please select type from the drop-down-list')
        for run_path, packet_path in zip(self.all_run_names, self.all_packets_names):
            try:
                upload_success = upload_to_cloud_api(batch_name=os.path.dirname(run_path),
                                                     tester_type=TesterType[self.tester_type],
                                                     run_data_csv_name=os.path.basename(run_path),
                                                     packets_data_csv_name=os.path.basename(packet_path),
                                                     env=self.values['env'],
                                                     is_batch_name_inside_logs_folder=False,
                                                     logger_=self.logger.name, owner_id=self.values['owner_id'])
            except Exception as e:
                self.logger.warning(f'UPLOAD: during upload_to_cloud_api an error occurred: {e}')
                upload_success = False

            if not upload_success:
                msg = "Run upload failed. Check exception error at the console and check Internet connection is " \
                      "available and upload logs manually"
                bg = 'red'
                popup_message(msg=msg, bg=bg)

            message = create_summary_message(is_upload=upload_success, data_folder_name=os.path.dirname(run_path))
            self.logger.info(message)
            all_status.append(upload_success)

        return all(all_status)


if __name__ == '__main__':
    # upload to cloud
    upload_obj = UploadTestersData()
    status = upload_obj.upload_to_cloud()

    print(f"done with status {status}")



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from wiliot_tools.utils.wiliot_gui.wiliot_gui import WiliotGui, popup_message
from wiliot_tools.test_equipment.test_equipment import YoctoSensor
from wiliot_tools.resolver_tool.resolve_packets import ResolvePackets

from wiliot_testers.tester_utils import dict_to_csv
import matplotlib
import matplotlib.pyplot as plt
from functools import partial
from wiliot_testers.utils.get_version import get_version
from wiliot_testers.utils.upload_to_cloud_api import upload_to_cloud_api
from wiliot_testers.yield_tester.modules.adva_process import *
from wiliot_testers.yield_tester.modules.count_thread import *
from wiliot_testers.yield_tester.utils.resolve_utils import *

import time

CONFIG_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'configs', 'gui_input_do_not_delete.json')
DEFAULT_USER_INPUT = {
    'inlay_dict_inlay': '', 'number': '', 'received_channel': '',
    'energy_pattern_val': '', 'tester_station_name': '',
    'comments': '', 'operator': '', 'wafer_lot': '', 'wafer_num': '',
    'conversion_type': '', 'surface': '', 'matrix_tags': '',
    'thermodes_col': 1, 'gw_energy_pattern': '', 'gw_time_profile': '',
    'window_size': 1, 'assembled_reel': '', 'do_resolve': False, 'owner_id': '',
}
RED_COLOR = 'red'
BLACK_COLOR = 'black'
SET_VALUE_MORE_THAN_100 = 110
VALUE_WHEN_NO_SENSOR = -10000
MIN_Y_FOR_PLOTS = 0
MAX_Y_FOR_PLOTS = 112
FIRST_STEP_SIZE = 10
TIME_BETWEEN_MATRICES = 0.5
RSSI_THRESHOLD = int(user_inputs.get('rssi_threshold', '0') or '0')

MAND_FIELDS = ['wafer_lot', 'wafer_num', 'window_size', 'thermodes_col', 'lanes',
               'q_size']  # mandatory fields in GUI before the run

matplotlib.use('TkAgg')
today = datetime.date.today()
formatted_today = today.strftime("%Y%m%d")  # without -
formatted_date = today.strftime("%Y-%m-%d")
current_time = datetime.datetime.now()
cur_time_formatted = current_time.strftime("%H%M%S")  # without :
time_formatted = current_time.strftime("%H:%M:%S")
logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s', datefmt='%m/%d/%Y %H:%M:%S %p')
root_logger = logging.getLogger()

for handler in root_logger.handlers:
    if isinstance(handler, logging.StreamHandler):
        handler.setLevel(logging.INFO)


class MainWindow:
    """
    The main class the runs the GUI and supervise the multi-threading process of fraction's calculation and GUI viewing
    """

    def __init__(self):
        self.curr_ex_id_for_log = None
        self.user_input = {}
        self.resolve_path = ''
        self.external_ids = []
        self.resolver = None
        self.resolve_q = None
        self.previous_input = {}
        self.current_values = None
        self.current_status_text = None
        self.cumulative_status_text = None
        self.assembled_reel = None
        self.lane_ids = ''
        self.light_intensity = None
        self.humidity = None
        self.temperature = None
        self.main_sensor = None
        self.main_gui = None
        self.test_started = True
        self.user_response_after_arduino_connection_error = False
        self.advanced_window = None
        self.user_response_after_gw_connection_error = False
        self.env_choice = 'prod'
        self.matrix_size = None
        self.filling_missed_field = None
        self.start_run = None
        self.logger = None
        self.ttfp = None
        self.curr_adva_for_log = None
        self.conversion_type = None
        self.surface = None
        self.adva_process = None
        self.adva_process_thread = None
        self.count_process = None
        self.count_process_thread = None
        self.resolver_thread = None
        self.folder_path = None
        self.py_wiliot_version = None
        self.final_path_run_data = None
        self.run_data_dict = None
        self.stop = threading.Event()
        self.thermodes_col = None
        self.print_neg_advas = True
        self.selected = ''
        self.wafer_lot = ''
        self.wafer_number = ''
        self.window_size = 1
        self.operator = ''
        self.tester_type = 'yield'
        self.tester_station_name = ''
        self.comments = ''
        self.gw_energy_pattern = None
        self.gw_time_profile = None
        self.rows_number = 1
        self.upload_flag = True
        self.cmn = ''
        self.final_path_packets_data = None
        self.seen_advas = set()
        self.not_neg_advas = 0  # used only to be shown in the small window
        self.update_packet_data_flag = False
        self.advas_before_tags = set()
        self.stop_run = False
        self.fig_canvas_agg1 = None
        self.trigger_count = 0
        self.yield_type = None

    def setup_logger(self, yield_tester_type):
        # Logger setup
        self.init_file_path(yield_tester_type)
        self.logger = logging.getLogger('yield')
        if self.logger.hasHandlers():
            self.logger.handlers.clear()
        self.logger.setLevel(logging.INFO)
        final_path_log_file = os.path.join(self.folder_path, self.cmn + '@yield_log.log')
        file_handler = logging.FileHandler(final_path_log_file)
        file_handler.setFormatter(
            logging.Formatter('%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p'))
        file_handler.setLevel(logging.INFO)
        self.logger.addHandler(file_handler)

    def get_adva_yield_result(self):
        """
        Calculates the yield fraction
        """
        result = 0
        tags_num = self.get_number_of_tested()
        if tags_num > 0:
            result = (self.not_neg_advas / tags_num) * 100
        return result

    def get_ex_yield_result(self):
        """
        Calculates the yield fraction
        """
        result = 0
        tags_num = self.get_number_of_tested()
        if tags_num > 0:
            result = (len(self.external_ids) / tags_num) * 100
        return result

    def run(self):
        """
        Viewing the window and checking if the process stops
        """
        self.open_session()
        if self.start_run:
            self.init_processes(self.selected)
            time.sleep(0.5)
            self.init_run_data()
            self.init_resolver_data()
            self.start_processes()
            self.overlay_window()
        else:
            self.logger.warning('Error Loading Program')

    def init_file_path(self, yield_tester_type):
        self.py_wiliot_version = get_version()
        d = WiliotDir()
        d.create_tester_dir(tester_name=yield_tester_type)
        yield_test_app_data = d.get_tester_dir(yield_tester_type)
        run_path = os.path.join(yield_test_app_data, self.get_log_folder_name())
        if not os.path.exists(run_path):
            os.makedirs(run_path)
        self.cmn = self.get_cmn(formatted_today, cur_time_formatted)
        self.folder_path = os.path.join(run_path, self.cmn)
        if not os.path.exists(self.folder_path):
            os.makedirs(self.folder_path)

    def get_log_folder_name(self):
        return self.wafer_lot + '.' + self.wafer_number

    def get_cmn(self, day_str, time_str):
        return f'{self.wafer_lot}.{self.wafer_number}_{day_str}_{time_str}'

    def init_resolver_data(self):
        if not self.user_input['do_resolve']:
            return

        self.resolve_path = os.path.join(self.folder_path, self.cmn + '@resolve_data.csv')
        dict_to_csv({'adv_address': [], 'resolve_status': [], 'external_id': []},
                    self.resolve_path, only_titles=True)

    def init_run_data(self):
        self.final_path_run_data = os.path.join(self.folder_path, self.cmn + '@run_data.csv')
        gw_version = self.adva_process.gw_instance.get_gw_version()[0]
        start_time = datetime.datetime.now()
        run_start_time = start_time.strftime("%H:%M:%S")
        value = inlays[self.selected]
        self.run_data_dict = {'common_run_name': self.cmn, 'tester_station_name': self.tester_station_name,
                              'operator': self.operator, 'received_channel': value['received_channel'],
                              'run_start_time': formatted_date + ' ' + run_start_time, 'run_end_time': '',
                              'wafer_lot': self.wafer_lot, 'wafer_number': self.wafer_number,
                              'assembled reel': self.assembled_reel, 'lane_ids': self.lane_ids,
                              'window_size': self.window_size, 'upload_date': '',
                              'tester_type': self.tester_type, 'gw_energy_pattern': self.gw_energy_pattern,
                              'comments': self.comments, 'inlay': self.selected, 'total_run_tested': 0,
                              'total_run_responding_tags': 0, 'conversion_type': self.conversion_type,
                              'gw_version': gw_version, 'surface': self.surface, 'matrix_tags': str(self.matrix_size),
                              'py_wiliot_version': self.py_wiliot_version, 'number_of_columns': self.thermodes_col,
                              'number_of_lanes': self.rows_number, 'gw_time_profile': self.gw_time_profile}

    def update_run_data_file(self, run_end_time, tags_num, advas, result, upload_date=''):

        """
        Updates the run_data CSV file while running the program
        """
        self.run_data_dict['run_end_time'] = run_end_time
        self.run_data_dict['upload_date'] = upload_date
        self.run_data_dict['total_run_tested'] = tags_num
        self.run_data_dict['total_run_responding_tags'] = advas
        self.run_data_dict['yield'] = result
        self.run_data_dict['conversion_type'] = self.conversion_type
        self.run_data_dict['surface'] = self.surface
        if self.user_input['do_resolve']:
            self.run_data_dict['total_run_external_ids'] = len(self.external_ids)
            self.run_data_dict['yield_external_ids'] = 100 * (len(self.external_ids) / tags_num) if tags_num else 0
        dict_to_csv(dict_in=self.run_data_dict, path=self.final_path_run_data)

    def calculate_ttfp_with_queue(self, packet_time, trigger_time_queue):
        while trigger_time_queue:
            try:
                trigger_time, matrix_cnt = trigger_time_queue[0]
                if len(trigger_time_queue) > 1:
                    trigger_time_next, _ = trigger_time_queue[1]
                else:
                    trigger_time_next = float('inf')
                # [matrix_x, time_x] <= p_time <= [matrix_y, time_y] --> p from matrix x
                if trigger_time <= packet_time <= trigger_time_next:
                    return trigger_time, matrix_cnt, packet_time - trigger_time
                elif packet_time < trigger_time:  # got packet before trigger
                    return float('nan'), matrix_cnt - 1, float('nan')
                else:  # packet_time > trigger_time_next
                    # Remove the processed trigger time
                    if ARDUINO_EXISTS:
                        self.count_process.pop_from_triggers_queue()
                    else:
                        self.adva_process.pop_from_triggers_queue()

            except Exception as e:
                self.logger.warning(f'Could not calculate tag matrix TTFP due to {e}')
                break
        return float('nan'), None, float('nan')

    def add_to_resolve_queue(self, packet_in):
        adva = packet_in.get_adva()
        if self.resolve_q.full():
            self.logger.warning(f'resolve queue is full discard the following adva: {adva}')
            return
        self.resolve_q.put({'tag': adva, 'payload': packet_in.get_payload()})

    def update_packet_data(self):
        """
        Updates the run_data CSV file while running the program
        """
        raw_packet_queue = self.adva_process.get_raw_packets_queue()
        triggers_queue = self.adva_process.get_trigger_time_queue() if not ARDUINO_EXISTS else \
            self.count_process.get_trigger_times()

        if not raw_packet_queue.empty():
            cur_df = pd.DataFrame()
            n_elements = raw_packet_queue.qsize()
            # Collecting Packets from the queue and putting them into a TagCollection
            for _ in range(n_elements):
                for p in raw_packet_queue.get():
                    trigger_time, matrix_cnt, tag_matrix_ttfp = self.calculate_ttfp_with_queue(p['time'],
                                                                                               triggers_queue)
                    cur_p = Packet(p['raw'], time_from_start=p['time'], inlay_type=self.selected,
                                   custom_data={
                                       'common_run_name': self.cmn,
                                       'matrix_tags_location': matrix_cnt,
                                       'matrix_timestamp': trigger_time,
                                       'tag_matrix_ttfp': tag_matrix_ttfp,
                                       'environment_light_intensity': self.light_intensity,
                                       'environment_humidity': self.humidity,
                                       'environment_temperature': self.temperature})

                    tag_id = cur_p.get_adva()

                    if self.get_number_of_tested() == 0:
                        self.advas_before_tags.add(tag_id)
                    else:
                        if self.print_neg_advas:
                            self.logger.info('neglected advas:  %05d', len(self.advas_before_tags))
                            self.print_neg_advas = False

                    if tag_id not in self.seen_advas and tag_id not in self.advas_before_tags:
                        cur_p_df = cur_p.as_dataframe(sprinkler_index=0)
                        cur_df = pd.concat([cur_df, cur_p_df], ignore_index=True)
                        self.seen_advas.add(tag_id)
                        self.logger.info(f"New adva {tag_id}")
                        if self.user_input['do_resolve']:
                            self.add_to_resolve_queue(cur_p)

            # writing to DataFrame and then to CSV
            if not cur_df.empty:
                self.final_path_packets_data = os.path.join(self.folder_path, f"{self.cmn}@packets_data.csv")
                try:
                    if not self.update_packet_data_flag:
                        cur_df.to_csv(self.final_path_packets_data, mode='w', header=True, index=False)
                        self.update_packet_data_flag = True
                    else:
                        cur_df.to_csv(self.final_path_packets_data, mode='a', header=False, index=False)
                except Exception as ee:
                    self.logger.error(f"Exception occurred: {ee}")

    def stop_button(self, run_end_time, tags_num, advas, result, upload_date):
        """
        Finishing the program and saves the last changes after pressing Stop in the second window
        """
        self.logger.info(f"User quit from application")
        self.adva_process_thread.join()
        self.update_run_data_file(formatted_date + ' ' + run_end_time, tags_num, advas, result, upload_date)
        self.update_packet_data()
        if ARDUINO_EXISTS:
            self.count_process_thread.join()
        if self.resolver_thread is not None:
            self.resolver_thread.join()

    def init_processes(self, inlay_select, time_between_matrices=TIME_BETWEEN_MATRICES, rssi_th=RSSI_THRESHOLD):
        """
        Initializing the two main instances and threads in order to start working
        """
        try:
            self.adva_process = AdvaProcess(stop_event=self.stop,
                                            inlay_type=inlay_select,
                                            logging_file=self.logger,
                                            listener_path=self.folder_path,
                                            time_between_matrices=time_between_matrices, rssi_th=rssi_th)
            self.adva_process_thread = threading.Thread(target=self.adva_process.run, args=())
        except Exception as e:
            self.logger.warning(f"{e}")
            popup_message(msg='GW is not connected. Please connect it.', logger=self.logger)
            raise Exception('GW is not connected')

        if ARDUINO_EXISTS:
            try:
                self.count_process = CountThread(self.stop, self.logger, self.matrix_size, self.thermodes_col)
                self.count_process_thread = threading.Thread(target=self.count_process.run, args=())
            except Exception as e:
                self.logger.warning(f"{e}")
                popup_message(msg='Arduino is not connected. Please connect it.', logger=self.logger)
                raise Exception('Arduino is not connected')

        if self.user_input['do_resolve']:
            self.yield_type = 'External Ids'
            resolve_q = Queue(maxsize=10000)
            self.resolver = ResolvePackets(tags_in_test=[],
                                           owner_id=self.user_input['owner_id'],
                                           env=ENV_RESOLVE,
                                           resolve_q=resolve_q,
                                           set_tags_status_df=self.updated_resolved_tags,
                                           stop_event_trig=self.stop,
                                           logger_name=self.logger.name,
                                           gui_type='ttk',
                                           tag_status=YieldTagStatus
                                           )
            self.resolver_thread = threading.Thread(target=self.resolver.run, args=())
            self.resolve_q = resolve_q
        else:
            self.yield_type = 'Advas'

    def updated_resolved_tags(self, tag_status):
        """
        tag_status = {'adv_address': [tag],
                      'resolve_status': [status],
                      'external_id': [ex_id]}
        @return:
        @rtype:
        """
        tag_status = {k: v[0] for k, v in tag_status.items()}
        new_ex_id = tag_status['external_id']
        self.logger.info(f'update resolved tags: {new_ex_id}')
        if new_ex_id not in self.external_ids:
            self.external_ids.append(new_ex_id)
        dict_to_csv(dict_in=tag_status, path=self.resolve_path, append=True)

    def start_processes(self):
        """
        Starting the work of the both threads
        """
        self.adva_process_thread.start()
        if ARDUINO_EXISTS:
            self.count_process_thread.start()
        if self.resolver_thread is not None:
            self.resolver_thread.start()

    @staticmethod
    def draw_figure(canvas, figure):
        """
        Embeds a Matplotlib figure in a PySimpleGUI Canvas Element
        """
        figure_canvas_agg = FigureCanvasTkAgg(figure, canvas)
        figure_canvas_agg.draw()
        figure_canvas_agg.get_tk_widget().grid(row=3, column=0, sticky="nsew", columnspan=120)
        figure_canvas_agg.get_tk_widget().configure(width=520, height=500)
        return figure_canvas_agg

    def upload_to_cloud(self):
        yes_or_no = ['Yes', 'No']
        upload_layout_dic = {'ask_to_upload': {'widget_type': 'label', 'value': 'Do you want to stop or upload?'},
                             'upload': {'text': 'Upload:', 'value': yes_or_no[0], 'widget_type': 'combobox',
                                        'options': yes_or_no},
                             'env_choice': {'text': 'Select Environment:', 'value': 'prod', 'widget_type': 'combobox',
                                            'options': ['prod', 'test']}}
        upload_layout_dic_gui = WiliotGui(params_dict=upload_layout_dic, parent=self.main_gui.layout)
        upload_layout_dic_values_out = upload_layout_dic_gui.run()

        if upload_layout_dic_values_out:
            self.upload_flag = upload_layout_dic_values_out['upload'] == 'Yes'
            self.env_choice = upload_layout_dic_values_out['env_choice']

        if self.upload_flag:
            try:
                is_uploaded = upload_to_cloud_api(self.cmn, self.tester_type + '-test',
                                                  run_data_csv_name=self.final_path_run_data,
                                                  packets_data_csv_name=self.final_path_packets_data,
                                                  env=self.env_choice, is_path=True)

            except Exception as ee:
                is_uploaded = False
                self.upload_flag = is_uploaded
                self.logger.error(f"Exception occurred: {ee}")
                exit()

            if is_uploaded:
                self.logger.info("Successful upload")
            else:
                self.logger.info('Failed to upload the file')
                popup_message(msg="Run upload failed. Check exception error at the console"
                                  " and check Internet connection is available"
                                  " and upload logs manually", tk_frame=self.main_gui.layout, logger=self.logger)
            self.main_gui.on_close()
            self.upload_flag = is_uploaded
        else:
            self.logger.info('File was not uploaded')

    def error_popup(self, error_type):
        self.logger.warning(f'{error_type} connection error occurred')
        popup_message(msg=f'{error_type} Connection error occurred.\n' f'Yield test was stopped',
                      tk_frame=self.main_gui.layout, logger=self.logger)
        self.logger.info(f'User reacted to {error_type} connection error')

    def get_current_plot_title(self):
        yield_type = 'Advas Yield' if self.yield_type == 'Advas' else 'External Ids Yield'

        return f'{yield_type} of last {self.window_size} matrices'

    def init_graphs(self, gui, min_current, min_cumulative):
        # create the main figure and two subplots
        fig, (ax, axy) = plt.subplots(1, 2, figsize=(12, 7))
        fig.subplots_adjust(left=0.05, right=0.95, top=0.85, bottom=0.15, wspace=0.2)
        # initialize the first graph
        prev_tests = None
        prev_val = None
        ax.set_xlabel('Number of tags')
        ax.set_ylabel('Yield %')
        ax.set_ylim([-MIN_Y_FOR_PLOTS, MAX_Y_FOR_PLOTS])
        ax_y_ticks = np.arange(MIN_Y_FOR_PLOTS, MAX_Y_FOR_PLOTS + 10, FIRST_STEP_SIZE)
        ax.set_yticks(ax_y_ticks)
        plt.ion()
        ax.yaxis.grid(True)
        text_box = ax.text(0.18, 1.05,
                           f"{self.get_current_plot_title()}: 0.00 %",
                           transform=ax.transAxes, fontweight='bold')
        if user_inputs.get('min_current_line') == 'yes':
            ax.axhline(y=min_current, color='black', linestyle='--')
        # initialize the second graph
        prev_tests1 = None
        prev_val1 = None
        axy.set_xlabel('Number of tags')
        axy.set_ylabel('Yield %')
        axy.set_ylim([MIN_Y_FOR_PLOTS, MAX_Y_FOR_PLOTS])
        axy_y_ticks = np.arange(MIN_Y_FOR_PLOTS, MAX_Y_FOR_PLOTS + 10, FIRST_STEP_SIZE)
        axy.set_yticks(axy_y_ticks)
        plt.ion()
        axy.yaxis.grid(True)
        text_box1 = axy.text(0.18, 1.05, f"Cumulative Yield: 0.0 %", transform=axy.transAxes,
                             fontweight='bold')
        if prev_val:
            text_box1 = axy.text(0.18, 1.05, f"Cumulative Yield: {prev_val:.2f} %", transform=axy.transAxes,
                                 fontweight='bold')
        if user_inputs.get('min_cumulative_line') == 'yes':
            axy.axhline(y=min_cumulative, color='black', linestyle='--')

        canvas_elem1 = gui.layout
        self.fig_canvas_agg1 = self.draw_figure(canvas_elem1, fig)
        return ax, axy, prev_tests, prev_val, prev_tests1, prev_val1, text_box, text_box1

    def calc_current_value(self):
        if self.yield_type == 'Advas':
            return 100 * ((len(self.seen_advas) - self.curr_adva_for_log) / self.matrix_size)
        else:
            return 100 * ((len(self.external_ids) - self.curr_ex_id_for_log) / self.matrix_size)

    def update_params_for_current_graph(self):
        self.new_trigger_log_line(latest_adva=len(self.seen_advas) - self.curr_adva_for_log,
                                  size=self.matrix_size)
        self.curr_adva_for_log = len(self.seen_advas)
        self.curr_ex_id_for_log = len(self.external_ids)

    def update_current_graph(self, new_trigger_graph_update):
        curr_tests = new_trigger_graph_update.get('new_num_rows')
        curr_val = self.calc_current_value()
        if curr_val > 100:
            curr_val = SET_VALUE_MORE_THAN_100

        if not hasattr(self, 'current_status_text') or self.current_status_text is None:
            self.current_status_text = None
        if curr_val < new_trigger_graph_update.get('min_current'):
            status_message_current = f'Current yield is lower than {new_trigger_graph_update.get("min_current")}%'
            if self.current_status_text is None:
                self.current_status_text = \
                    new_trigger_graph_update.get('ax').text(0.5, -0.19, status_message_current,
                                                            transform=new_trigger_graph_update.get('ax').transAxes,
                                                            fontsize=12, color='red', fontweight='bold', ha='center')
            else:
                self.current_status_text.set_text(status_message_current)
        else:
            if self.current_status_text is not None:
                self.current_status_text.remove()
                self.current_status_text = None

        figure_color_current = 'red' if curr_val < new_trigger_graph_update.get('red_line_current') else 'blue'
        # Plot the first point only if it's the first update
        new_trigger_graph_update.get('ax').plot([new_trigger_graph_update.get('prev_tests'), curr_tests],
                                                [new_trigger_graph_update.get('prev_val'), curr_val],
                                                color=figure_color_current)
        prev_tests = curr_tests
        prev_val = curr_val
        new_trigger_graph_update.get('text_box').set_text(f"{self.get_current_plot_title()}: {curr_val} %")

        self.update_params_for_current_graph()

        return prev_tests, prev_val

    def update_cumulative_graph(self, cumulative_graph_update_args):
        curr_tests1 = cumulative_graph_update_args.get('new_num_rows')
        curr_val1 = self.get_adva_yield_result() if self.yield_type == 'Advas' else self.get_ex_yield_result()

        if curr_val1 > 100:
            curr_val1 = SET_VALUE_MORE_THAN_100

        if not hasattr(self, 'cumulative_status_text') or self.cumulative_status_text is None:
            self.cumulative_status_text = None
        if curr_val1 < cumulative_graph_update_args.get('min_cumulative') and self.get_number_of_tested() != 0:
            status_message_cumulative = \
                f'Cumulative yield is lower than {cumulative_graph_update_args.get("min_cumulative")}%'
            if self.cumulative_status_text is None:
                self.cumulative_status_text = cumulative_graph_update_args.get('axy').text(
                    0.5, -0.19, status_message_cumulative,
                    transform=cumulative_graph_update_args.get('axy').transAxes,
                    fontsize=12, color='red', fontweight='bold', ha='center')
            else:
                self.cumulative_status_text.set_text(status_message_cumulative)
        else:
            if self.cumulative_status_text is not None:
                self.cumulative_status_text.remove()
                self.cumulative_status_text = None
        figure_color_cumulative = 'red' if curr_val1 < cumulative_graph_update_args.get(
            'red_line_cumulative') else 'blue'
        cumulative_graph_update_args.get('axy').plot([cumulative_graph_update_args.get('prev_tests1'), curr_tests1],
                                                     [cumulative_graph_update_args.get('prev_val1'), curr_val1],
                                                     color=figure_color_cumulative)
        prev_tests1 = curr_tests1
        prev_val1 = curr_val1
        yield_type_str = 'Advas' if self.yield_type == 'Advas' else 'External Ids'
        cumulative_graph_update_args.get('text_box1').set_text(f"{yield_type_str} Cumulative Yield: {curr_val1:.2f}%")
        self.fig_canvas_agg1.draw()

        return prev_tests1, prev_val1

    def handling_advanced_settings_window(self, advanced_settings_inputs):
        self.logger.info('Advanced settings was pressed')
        advanced_layout_dict = {
            'current_min_y_value': {'text': 'Current" min y:', 'value': '', 'widget_type': 'entry'},
            'current_max_y_value': {'text': '"Current" max y:', 'value': '', 'widget_type': 'entry'},
            'current_size_value': {'text': '"Current" step:', 'value': '', 'widget_type': 'entry'},
            'cumulative_min_y_value': {'text': '"Cumulative" min y:', 'value': '', 'widget_type': 'entry'},
            'cumulative_max_y_value': {'text': '"Cumulative" max y', 'value': '', 'widget_type': 'entry'},
            'cumulative_size_value': {'text': '"Cumulative" step:', 'value': '', 'widget_type': 'entry'},
            'window_size': {'text': 'Window size:', 'value': '', 'widget_type': 'entry'},
            'yield_type': {'text': 'Yield type', 'value': self.yield_type, 'widget_type': 'combobox',
                           'options': ['Advas', 'External Ids']},
            'reset_button': {'text': 'Reset', 'value': '', 'widget_type': 'button'},
        }

        default_values = {
            'current_min_y_value': 0,
            'current_max_y_value': 120,
            'current_size_value': 10,
            'cumulative_min_y_value': 0,
            'cumulative_max_y_value': 120,
            'cumulative_size_value': 10,
            'window_size': 1,
            'yield_type': 'Advas',

        }

        def reset_button():
            self.logger.info("Reset values from advanced settings")
            for key in default_values:
                advanced_gui.update_widget(key, default_values[key])
                globals()[key] = default_values[key]

            advanced_settings_inputs.get('ax').set_ylim([advanced_settings_inputs.get('current_min_y_value'),
                                                         advanced_settings_inputs.get('current_max_y_value')])
            advanced_settings_inputs.get('ax').set_yticks(
                np.arange(advanced_settings_inputs.get('current_min_y_value'),
                          advanced_settings_inputs.get('current_max_y_value') + advanced_settings_inputs.get(
                              'current_size_value'),
                          advanced_settings_inputs.get('current_size_value')))

            advanced_settings_inputs.get('axy').set_ylim([advanced_settings_inputs.get('cumulative_min_y_value'),
                                                          advanced_settings_inputs.get('cumulative_max_y_value')])
            advanced_settings_inputs.get('axy').set_yticks(
                np.arange(advanced_settings_inputs.get('cumulative_min_y_value'),
                          advanced_settings_inputs.get('cumulative_max_y_value') + advanced_settings_inputs.get(
                              'cumulative_size_value'),
                          advanced_settings_inputs.get('cumulative_size_value')))
            self.main_gui.window_size = 1
            self.yield_type = 'Advas'

        advanced_gui = WiliotGui(params_dict=advanced_layout_dict, exit_sys_upon_cancel=False,
                                 parent=self.main_gui.layout)
        advanced_gui.set_button_command('reset_button', reset_button)
        user_out = advanced_gui.run()

        if not self.current_values:
            self.current_values = {
                'current_min_y_value': advanced_settings_inputs.get('current_min_y_value'),
                'current_max_y_value': advanced_settings_inputs.get('current_max_y_value'),
                'current_size_value': advanced_settings_inputs.get('current_size_value'),
                'cumulative_min_y_value': advanced_settings_inputs.get('cumulative_min_y_value'),
                'cumulative_max_y_value': advanced_settings_inputs.get('cumulative_max_y_value'),
                'cumulative_size_value': advanced_settings_inputs.get('cumulative_size_value'),
                'window_size': advanced_settings_inputs.get('window_size', 1),
                'yield_type': 'Advas',
            }

        def get_adv_value():
            if not adv_value.isdigit() and adv_value != '':
                popup_message(msg=f"A not-number character in {adv_key}", tk_frame=advanced_gui.layout,
                              logger=self.logger)
                return
            if adv_value != '':
                new_value = int(adv_value)  # if user wrote a value
            elif self.current_values[adv_key]:
                new_value = self.current_values[
                    adv_key]  # if the user did not write a value and not the first submit
            else:
                new_value = int(default_values[adv_key])  # if the user did not write a value and first submit
            return new_value

        if user_out:
            self.main_gui.layout.attributes('-alpha', 1.0)
            for adv_key, adv_value in user_out.items():
                if adv_value and adv_key != 'window_size' and adv_key != 'yield_type':
                    new_value = get_adv_value()
                else:
                    if adv_key == 'yield_type':
                        new_value = adv_value
                        self.current_values[adv_key] = new_value
                        if self.yield_type != new_value:
                            for line in advanced_settings_inputs.get('ax').get_lines():
                                line.remove()
                            for line in advanced_settings_inputs.get('axy').get_lines():
                                line.remove()
                            self.yield_type = new_value
                    else:
                        new_value = int(get_adv_value())
                        self.window_size = new_value
                        self.main_gui.window_size = new_value
                self.current_values[adv_key] = new_value
                self.logger.info(f"{adv_key} changed to {new_value}")
            advanced_settings_inputs.get('ax').set_ylim(
                [self.current_values['current_min_y_value'], self.current_values['current_max_y_value']])
            advanced_settings_inputs.get('ax').set_yticks(np.arange(self.current_values['current_min_y_value'],
                                                                    self.current_values['current_max_y_value'] +
                                                                    self.current_values[
                                                                        'current_size_value'],
                                                                    self.current_values['current_size_value']))

            advanced_settings_inputs.get('axy').set_ylim(
                [self.current_values['cumulative_min_y_value'], self.current_values['cumulative_max_y_value']])
            advanced_settings_inputs.get('axy').set_yticks(np.arange(self.current_values['cumulative_min_y_value'],
                                                                     self.current_values['cumulative_max_y_value'] +
                                                                     self.current_values[
                                                                         'cumulative_size_value'],
                                                                     self.current_values['cumulative_size_value']))
            advanced_settings_inputs.get('text_box').set_text(
                f"{self.get_current_plot_title()}: {self.calc_current_value()} %")
            self.fig_canvas_agg1.draw()

    def get_number_of_tested(self):
        if ARDUINO_EXISTS:
            tags_num = self.count_process.get_tested()
        else:
            tags_num = self.adva_process.get_sensors_triggers() * int(self.thermodes_col) * int(self.rows_number)
        return tags_num

    def stop_or_error_process(self):
        end_time = datetime.datetime.now()
        run_end_time = end_time.strftime("%H:%M:%S")
        if self.upload_flag:
            upload_date = run_end_time
        else:
            upload_date = ''
        advas = len(self.seen_advas)
        tags_num = self.get_number_of_tested()
        result = float(100 * (advas / tags_num)) if tags_num != 0 else float('inf')
        self.stop_button(run_end_time, tags_num, advas, result, upload_date)
        return True

    def new_trigger_log_line(self, latest_adva, size):
        yield_result = "%.5f" % self.get_adva_yield_result()

        if '.' in yield_result and len(yield_result.split('.')[0]) < 2:
            yield_result = "0" + yield_result
        latest_yield_formatted = "{:.5f}".format(float(latest_adva / size) * 100).zfill(9)
        if ARDUINO_EXISTS:
            matrix_num = self.count_process.get_tested() / size
            all_tested = self.count_process.get_tested()
        else:
            matrix_num = self.adva_process.get_sensors_triggers()
            all_tested = self.adva_process.get_sensors_triggers() * size
        self.logger.info(
            'Matrix Number: %05d, Cumulative Yield: %s, Cumulative Tags: %05d, Cumulative Advas: %05d,'
            'Latest Yield: %s, Latest Tags: %05d, Latest Advas: %05d, Light Intensity: '
            '%05.1f, Humidity: %05.1f, Temperature: %05.1f',
            matrix_num, yield_result, all_tested, len(self.seen_advas), latest_yield_formatted,
            size, latest_adva, self.light_intensity, self.humidity, self.temperature)

    def new_trigger_sensors(self, new_trigger_user_configs):

        temperature_display = f"{self.temperature:.2f} °C"
        if self.main_sensor:
            self.light_intensity = self.main_sensor.get_light()
            self.humidity = self.main_sensor.get_humidity()
            self.temperature = self.main_sensor.get_temperature()
        if new_trigger_user_configs.get('temperature_type') == "F":
            temperature_display = f"{self.temperature * 9 / 5 + 32:.2f} °F"
        temperature_color = BLACK_COLOR if (new_trigger_user_configs.get('min_temperature') <= self.temperature <=
                                            new_trigger_user_configs.get('max_temperature')) else RED_COLOR
        light_intensity_color = BLACK_COLOR if (
                new_trigger_user_configs.get('min_light_intensity') <= self.light_intensity <=
                new_trigger_user_configs.get('max_light_intensity')) else RED_COLOR
        humidity_color = BLACK_COLOR if (new_trigger_user_configs.get('min_humidity') <= self.humidity <=
                                         new_trigger_user_configs.get('max_humidity')) else RED_COLOR
        self.main_gui.update_widget('sensor_row_temperature_value', f'Temperature: {temperature_display}',
                                    color=temperature_color)
        self.main_gui.update_widget('sensor_row_light_intensity_value',
                                    f'Light Intensity: {self.light_intensity} lux',
                                    color=light_intensity_color)
        self.main_gui.update_widget('sensor_row_humidity_value', f'Humidity: {self.humidity} %',
                                    color=humidity_color)

    def overlay_window(self):
        """
        The small window open session
        """
        # taking values from user_input json file
        temperature_type = user_inputs.get('temperature_type', default_user_inputs['temperature_type'])
        min_current = float(user_inputs.get('min_current', default_user_inputs['min_current']))
        min_cumulative = float(user_inputs.get('min_cumulative', default_user_inputs['min_cumulative']))
        min_humidity = float(user_inputs.get('min_humidity', default_user_inputs['min_humidity']))
        max_humidity = float(user_inputs.get('max_humidity', default_user_inputs['max_humidity']))
        max_light_intensity = float(user_inputs.get('max_light_intensity', default_user_inputs['max_light_intensity']))
        min_light_intensity = float(user_inputs.get('min_light_intensity', default_user_inputs['min_light_intensity']))
        min_temperature = float(user_inputs.get('min_temperature', default_user_inputs['min_temperature']))
        max_temperature = float(user_inputs.get('max_temperature', default_user_inputs['max_temperature']))
        red_line_current = float(user_inputs.get('red_line_current', default_user_inputs['red_line_current']))
        red_line_cumulative = float(user_inputs.get('red_line_cumulative', default_user_inputs['red_line_cumulative']))

        # creating the main window
        temp_val = self.temperature if temperature_type == "C" else self.temperature * 9 / 5 + 32
        overlay_layout_dict = {
            'counting_row': [{'num_rows': {'text': '', 'widget_type': 'label', 'value': 'Number of tags:',
                                           'options': {'font': ('Arial', 26, 'bold')}}},
                             {'num_advas': {'text': '', 'widget_type': 'label', 'value': 'Number of advas:',
                                            'options': {'font': ('Arial', 26, 'bold')}}},
                             ],

            'sensor_row': [{'light_intensity_value': {'text': '', 'widget_type': 'label',
                                                      'options': {'font': ('Arial', 14, 'bold')},
                                                      'value': f'Light Intensity: {self.light_intensity}'}},
                           {'temperature_value': {'widget_type': 'label', 'options': {'font': ('Arial', 14, 'bold')},
                                                  'value': f'Temperature: {temp_val} {temperature_type}'}},
                           {'humidity_value': {'widget_type': 'label', 'value': f'Humidity: {self.humidity}',
                                               'options': {'font': ('Arial', 14, 'bold')}, }
                            }, ],

            'space': {'value': '', 'widget_type': 'label'},
            'buttons_row': [
                {'advanced_settings_button': {'text': 'Advanced Settings', 'value': '', 'widget_type': 'button'}},
                {'stop_button': {'text': 'Stop', 'value': '', 'widget_type': 'button'}},
                {'pause_button': {'text': 'Pause Test', 'value': '', 'widget_type': 'button'}}]

        }
        if self.user_input['do_resolve']:
            overlay_layout_dict['counting_row'].append(
                {'num_ex_ids': {'text': '', 'widget_type': 'label', 'value': 'Number of external ids:',
                                'options': {'font': ('Arial', 26, 'bold')}}}
            )

        def stop_button_callback():
            self.stop.set()
            final_tags = self.get_number_of_tested()
            self.logger.info('Final Yield: %s, Final Tags: %05d, Final Advas: %05d,',
                             self.get_adva_yield_result(), final_tags, len(self.seen_advas), )
            self.upload_to_cloud()
            self.stop_run = self.stop_or_error_process()

        def toggle_test_callback():
            self.test_started = not self.test_started
            if ARDUINO_EXISTS:
                self.count_process.set_pause_triggers(not self.test_started)
            self.adva_process.set_stopped_by_user(not self.test_started)
            if self.test_started:
                self.logger.info('Test was started by user')
                self.main_gui.update_widget('buttons_row_pause_button', 'Pause Test')

            else:
                self.logger.info('Test was paused by user')
                self.main_gui.update_widget('buttons_row_pause_button', 'Start Test')

        # values before running
        self.curr_adva_for_log = len(self.seen_advas)
        self.curr_ex_id_for_log = len(self.external_ids)
        sub = False
        current_min_y_value = MIN_Y_FOR_PLOTS
        current_max_y_value = MAX_Y_FOR_PLOTS
        current_size_value = FIRST_STEP_SIZE
        cumulative_min_y_value = MIN_Y_FOR_PLOTS
        cumulative_max_y_value = MAX_Y_FOR_PLOTS
        cumulative_size_value = FIRST_STEP_SIZE
        result = float('inf')

        self.main_gui = WiliotGui(params_dict=overlay_layout_dict, full_screen=True, do_button_config=False)
        # initializing graphs
        ax, axy, prev_tests, prev_val, prev_tests1, prev_val1, text_box, text_box1 = \
            self.init_graphs(self.main_gui, min_current, min_cumulative)
        advanced_settings_inputs = {
            'ax': ax,
            'current_min_y_value': current_min_y_value,
            'current_max_y_value': current_max_y_value,
            'current_size_value': current_size_value,
            'axy': axy,
            'cumulative_min_y_value': cumulative_min_y_value,
            'cumulative_max_y_value': cumulative_max_y_value,
            'cumulative_size_value': cumulative_size_value,
            'window_size': self.window_size,
            'text_box': text_box,
        }
        advanced_settings_callback = partial(
            self.handling_advanced_settings_window, advanced_settings_inputs)

        self.main_gui.set_button_command('buttons_row_advanced_settings_button', advanced_settings_callback)
        self.main_gui.set_button_command('buttons_row_stop_button', stop_button_callback)
        self.main_gui.set_button_command('buttons_row_pause_button', toggle_test_callback)

        # initialize num_advas and num_rows
        num_rows = 0
        num_advas = 0
        self.main_gui.update_widget('counting_row_num_rows', f"Number of tags: {num_rows}")
        self.main_gui.update_widget('counting_row_num_advas', f"Number of advas: {num_advas}")
        if self.user_input['do_resolve']:
            self.main_gui.update_widget('counting_row_num_ex_ids', f"Number of external ids: {len(self.external_ids)}")

        def update_gui():
            nonlocal num_rows, num_advas, prev_tests, prev_val, prev_tests1, prev_val1, result, sub

            new_num_rows = self.get_number_of_tested()
            new_num_advas = len(self.seen_advas)
            self.not_neg_advas = new_num_advas

            # update packet data
            self.update_packet_data()

            if self.user_input['do_resolve']:
                self.main_gui.update_widget('counting_row_num_ex_ids',
                                            f"Number of external ids: {len(self.external_ids)}")

            # updating number of rows in GUI
            if new_num_rows != num_rows:
                num_rows = new_num_rows
                self.main_gui.update_widget('counting_row_num_rows', f"Number of tags: {num_rows}")

            # updating number of advas in GUI
            if new_num_advas != num_advas and new_num_advas > 0:
                num_advas = new_num_advas
                self.main_gui.update_widget('counting_row_num_advas', f"Number of advas: {num_advas}")

            # all processes when getting a new matrix
            curr_triggers_count = self.adva_process.get_sensors_triggers() if not ARDUINO_EXISTS \
                else self.count_process.get_matrix_cnt()
            triggers_diff = curr_triggers_count - self.trigger_count
            if triggers_diff > int(self.window_size) - 1:  # not using == or >=
                self.trigger_count = curr_triggers_count
                new_trigger_graph_update = {
                    'text_box': text_box,
                    'ax': ax,
                    'new_num_rows': new_num_rows,
                    'min_current': min_current,
                    'red_line_current': red_line_current,
                    'prev_tests': prev_tests,
                    'prev_val': prev_val,
                }
                new_trigger_user_configs = {
                    'temperature_type': temperature_type,
                    'min_temperature': min_temperature,
                    'max_temperature': max_temperature,
                    'min_light_intensity': min_light_intensity,
                    'max_light_intensity': max_light_intensity,
                    'min_humidity': min_humidity,
                    'max_humidity': max_humidity,

                }
                self.new_trigger_sensors(new_trigger_user_configs=new_trigger_user_configs)
                # updating the first graph
                prev_tests, prev_val = self.update_current_graph(new_trigger_graph_update)

            # updating the second graph
            cumulative_graph_args = {
                'axy': axy,
                'new_num_rows': new_num_rows,
                'min_cumulative': min_cumulative,
                'red_line_cumulative': red_line_cumulative,
                'prev_tests1': prev_tests1,
                'prev_val1': prev_val1,
                'text_box1': text_box1
            }
            prev_tests1, prev_val1 = self.update_cumulative_graph(cumulative_graph_update_args=cumulative_graph_args)

            end_time = datetime.datetime.now()
            run_end_time = end_time.strftime("%H:%M:%S")
            advas = len(self.seen_advas)
            tags_num = self.get_number_of_tested()
            result = float(100 * (advas / tags_num)) if tags_num != 0 else float('inf')
            self.update_run_data_file(formatted_date + ' ' + run_end_time, tags_num, advas, result)
            if self.adva_process.get_gw_error_connection() or \
                    (ARDUINO_EXISTS and self.count_process.get_arduino_connection_error()):
                if ARDUINO_EXISTS:
                    if self.count_process.get_arduino_connection_error():
                        self.user_response_after_arduino_connection_error = True
                        self.logger.warning('User responded to Arduino error')
                self.user_response_after_gw_connection_error = True
                self.logger.warning('User responded to GW error')
                self.stop_or_error_process()
            if self.user_response_after_gw_connection_error:
                connection_error = 'GW'
                self.error_popup(connection_error)
                self.upload_to_cloud()
                end_time = datetime.datetime.now()
                run_end_time = end_time.strftime("%H:%M:%S")
                if self.upload_flag:
                    upload_date = run_end_time
                else:
                    upload_date = ''
                self.update_run_data_file(formatted_date + ' ' + run_end_time, self.get_number_of_tested(),
                                          len(self.seen_advas), result, upload_date)
                self.main_gui.on_close()

            if sub or self.stop_run:
                self.stop_or_error_process()
                time.sleep(1)
                self.adva_process_thread.join()
                sys.exit()

        self.main_gui.add_recurrent_function(100, update_gui)
        self.main_gui.run()

    @staticmethod
    def open_session_layout(previous_input, inlay_info):

        open_session_layout = {
            'wafer_lot': {'text': 'Wafer Lot:', 'value': previous_input['wafer_lot'], 'widget_type': 'entry'},
            'wafer_num': {'text': 'Wafer Number:', 'value': previous_input['wafer_num'], 'widget_type': 'entry'},
            'thermodes_col': {'text': 'Number of Columns:', 'value': previous_input['thermodes_col'],
                              'widget_type': 'entry'},
            'matrix_tags': {'text': '', 'widget_type': 'label',
                            'value': f'Matrix tags: {str(inlay_info["default_matrix_tags"])}'},
            'inlay_dict': [
                {'inlay': {'text': 'Inlay:', 'value': previous_input['inlay_dict_inlay'], 'widget_type': 'combobox',
                           'options': list(inlays.keys())}},
                {'inlay_info': {'widget_type': 'label', 'value': inlay_info['inlay_info']}},
            ],
            'tester_station_name': {'text': 'Tester Station:', 'value': previous_input['tester_station_name'],
                                    'widget_type': 'entry'},
            'comments': {'text': 'Comments:', 'value': previous_input['comments'], 'widget_type': 'entry'},
            'operator': {'text': 'Operator:', 'value': previous_input['operator'], 'widget_type': 'entry'},
            'conversion_type': {'text': 'Conversion:', 'value': previous_input['conversion_type'],
                                'widget_type': 'combobox', 'options': inlay_info["conv_opts"]},
            'surface': {'text': 'Surface:', 'value': previous_input['surface'], 'widget_type': 'combobox',
                        'options': inlay_info["surfaces"]},
            'window_size': {'text': 'Window Size for Analysis:', 'value': previous_input['window_size'],
                            'widget_type': 'entry'},
            'do_resolve': {'text': 'Get External Id from Cloud', 'value': previous_input['do_resolve']},
            'owner_id': {'text': 'Owner Id for Cloud Connection', 'value': previous_input['owner_id']},

        }

        return open_session_layout

    def setup_inlay_parameters(self, values_out, energy_pat, time_pro, thermodes_col=None, rows_num=None,
                               yield_tester_type='assembly_yield_tester'):
        selected_inlay = inlays.get(self.selected)
        self.user_input = values_out
        self.rows_number = int(selected_inlay['number_of_rows']) if rows_num is None else rows_num
        self.wafer_lot = values_out.get('wafer_lot', '')
        self.wafer_number = values_out.get('wafer_num', '')
        self.window_size = values_out.get('window_size', 1)
        self.comments = values_out['comments']
        self.gw_energy_pattern = energy_pat
        self.gw_time_profile = time_pro
        self.thermodes_col = values_out.get('thermodes_col', 1) if thermodes_col is None else thermodes_col
        self.conversion_type = values_out['conversion_type']
        self.surface = values_out['surface']
        self.tester_station_name = values_out['tester_station_name']
        self.operator = values_out['operator']
        self.matrix_size = int(self.thermodes_col) * int(self.rows_number)
        self.setup_logger(yield_tester_type)

        try:
            self.main_sensor = YoctoSensor(self.logger)
        except Exception as ee:
            self.main_sensor = None
            print(f'No sensor is connected ({ee})')
        if self.main_sensor:
            self.light_intensity = self.main_sensor.get_light()
            self.humidity = self.main_sensor.get_humidity()
            self.temperature = self.main_sensor.get_temperature()
        else:
            self.temperature = VALUE_WHEN_NO_SENSOR
            self.humidity = VALUE_WHEN_NO_SENSOR
            self.light_intensity = VALUE_WHEN_NO_SENSOR

        self.start_run = True

    def open_session(self, cols_or_rows='thermodes_col'):
        """
        Opening a session for the process
        """
        # Load previous input from the config file
        if os.path.exists(CONFIG_FILE):
            with open(CONFIG_FILE, "r") as f:
                self.previous_input = json.load(f)

        for k, v in DEFAULT_USER_INPUT.items():
            if k not in self.previous_input.keys():
                self.previous_input[k] = v

        self.start_run = False
        self.selected = self.previous_input['inlay_dict_inlay']
        selected_inlay = inlays.get(self.selected, {})
        energy_pat = selected_inlay.get('energy_pattern_val', 'Invalid Selection')
        time_pro = selected_inlay.get('time_profile_val', 'Invalid Selection')
        default_matrix_tags = int(self.previous_input.get('thermodes_col', 1))

        inlay_info_dict = {
            'inlay_info': ',   '.join(f"{key}: {value}" for key, value in selected_inlay.items() if
                                      key != 'inlay' and key != 'number' and key != 'number_of_rows'),
            'default_matrix_tags': default_matrix_tags,
            'conv_opts': ['Not converted', 'Standard', 'Durable'],
            'surfaces': ['Air', 'Cardboard', 'RPC', 'General Er3', 'General Er3.5']
        }

        open_session_layout = self.open_session_layout(previous_input=self.previous_input, inlay_info=inlay_info_dict)

        open_session_gui = WiliotGui(params_dict=open_session_layout, do_button_config=False)

        def on_inlay_change(*args):
            inlay_var = open_session_gui.widgets_vals['inlay_dict_inlay']
            inlay_select = inlay_var.get()
            self.selected = inlay_select
            if inlay_select in inlays:
                selected_inlay = inlays[inlay_select]
                inlay_info = inlays[self.selected]
                info_string = ',   '.join(
                    f"{key}: {value}" for key, value in inlay_info.items() if
                    key != 'inlay' and key != 'number' and key != 'number_of_rows')
                thermodes_col_value = open_session_gui.widgets_vals.get('thermodes_col')
                if thermodes_col_value is not None:
                    default_matrix_tags = int(thermodes_col_value.get()) * selected_inlay.get('number_of_rows', 1)
                else:
                    default_matrix_tags = selected_inlay.get('number_of_rows', 1)
            else:
                info_string = 'Invalid Selection'
                default_matrix_tags = 0

            open_session_gui.update_widget('inlay_dict_inlay_info', info_string)
            if 'matrix_tags' in open_session_gui.widgets.keys():
                open_session_gui.update_widget('matrix_tags', f'Matrix tags: {str(default_matrix_tags)}')

        def on_cols_change(*args):
            nonlocal default_matrix_tags
            inlay_var = open_session_gui.widgets_vals['inlay_dict_inlay']
            inlay = inlay_var.get() if inlay_var.get() != '' and inlay_var.get() in inlays.keys() else list(inlays)[0]
            selected_inlay = inlays[inlay]
            cols_str = open_session_gui.widgets_vals.get('thermodes_col')
            rows_str = open_session_gui.widgets_vals.get('rows_num')
            cols_num = cols_str.get() if cols_str is not None else 1
            rows_num = int(rows_str.get()) if rows_str is not None else selected_inlay.get('number_of_rows')
            default_matrix_tags = cols_num * rows_num
            if 'matrix_tags' in open_session_gui.widgets.keys():
                open_session_gui.update_widget('matrix_tags', f'Matrix tags: {str(default_matrix_tags)}')
            if 'parts_row_part_1' in open_session_gui.widgets.keys():
                for i in range(1, 4):
                    open_session_gui.update_widget(f'parts_row_part_{i}', disabled=(i > rows_num))

        def submit_open_session(*args):
            wafer_number = ''
            wafer_lot = ''
            wafer_number_widget = open_session_gui.widgets.get('wafer_num')
            if wafer_number_widget is not None:
                wafer_number = wafer_number_widget.get()
            wafer_lot_widget = open_session_gui.widgets.get('wafer_lot')
            if wafer_lot_widget is not None:
                wafer_lot = wafer_lot_widget.get()
            missing_fields = []
            self.filling_missed_field = []
            for field in MAND_FIELDS:
                if open_session_gui.widgets.get(field) is not None:
                    session_value = open_session_gui.widgets.get(field).get().strip()
                    if not session_value:
                        missing_fields.append(field)
                        self.filling_missed_field.append(field)
            if missing_fields:
                error_msg = f"Please fill all the " \
                            f"mandatory fields {', '.join([f'[{field}]' for field in missing_fields])}"
                popup_message(msg=error_msg, tk_frame=open_session_gui.layout, logger=self.logger)
                return  # Skip the rest and prompt for missing fields again
            for missed_field in self.filling_missed_field:
                setattr(self, missed_field, open_session_gui.widgets.get(missed_field).get().strip())

            for value, value_name in [(wafer_lot, "Wafer lot"), (wafer_number, "Wafer number")]:
                for character in value:
                    if not character.isalpha() and not character.isdigit():
                        popup_message(msg=f"{value_name} can't include '{character}' not letter/digit",
                                      tk_frame=open_session_gui.layout, logger=self.logger)
                        return
            open_session_gui.on_submit()

        open_session_gui.button_configs(submit_command=submit_open_session)
        open_session_gui.add_event(widget_key='inlay_dict_inlay', command=on_inlay_change)
        open_session_gui.add_event(widget_key=cols_or_rows, command=on_cols_change)

        on_cols_change()
        values_out = open_session_gui.run(save_path=CONFIG_FILE)

        if values_out:
            self.setup_inlay_parameters(values_out, energy_pat, time_pro)
            # Save the correct values to previous_input
            # self.saving_prev_input(values_out)


if __name__ == '__main__':
    m = MainWindow()
    m.run()



all_inlays = {
    'TIKI_117': {'inlay': 'TIKI_117', 'number': '117', 'number_of_rows': 5, 'received_channel': '37',
                 'energy_pattern_val': 51, 'time_profile_val': [5, 40], 'symbol_val': '1Mhz'},
    'TIKI_121': {'inlay': 'TIKI_121', 'number': '121', 'number_of_rows': 6, 'received_channel': '37',
                 'energy_pattern_val': 51, 'time_profile_val': [5, 40], 'symbol_val': '1Mhz'},
    'TIKI_169': {'inlay': 'TIKI_169', 'number': '169', 'number_of_rows': 4, 'received_channel': '4',
                 'energy_pattern_val': 18, 'time_profile_val': [5, 40], 'symbol_val': '1Mhz'},
    'TIKI_169_68': {'inlay': 'TIKI_169', 'number': '169', 'number_of_rows': 4, 'received_channel': '0',
                    'energy_pattern_val': 18, 'time_profile_val': [5, 40], 'symbol_val': '1Mhz'},
    'TIKI_169_BLE5_C': {'inlay': 'TIKI_169', 'number': '169', 'number_of_rows': 4, 'received_channel': '10',
                        'energy_pattern_val': 18, 'time_profile_val': [5, 40], 'symbol_val': '2Mhz'},
}
